#  Copyright (c) 2024 zfit

from __future__ import annotations

from collections.abc import Callable
from typing import Union, Optional

import pydantic

from ..serialization import Serializer, SpaceRepr

from typing import Literal

import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp
from tensorflow_probability.python import distributions as tfd

import zfit.z.numpy as znp
from .dist_tfp import WrapDistribution
from .. import z
from ..core.basepdf import BasePDF
from ..core.interfaces import ZfitData, ZfitParameter, ZfitSpace
from ..core.serialmixin import SerializableMixin
from ..serialization.pdfrepr import BasePDFRepr
from ..settings import ztypes, run
from ..util import (
    binning as binning_util,
    convolution as convolution_util,
    improved_sheather_jones as isj_util,
    ztyping,
)
from ..util.exception import OverdefinedError, ShapeIncompatibleError
from ..util.ztyping import ExtendedInputType, NormInputType
from ..z.math import weighted_quantile


@z.function(wraps="tensor", keepalive=True)
def bandwidth_rule_of_thumb(
    data: znp.array,
    weights: znp.array | None,
    factor: float | int | znp.array = None,
) -> znp.array:
    r"""Calculate the bandwidth of *data* using a rule of thumb.

    This calculates a global, single bandwidth for all kernels using a rule of thumb.
    |@doc:pdf.kde.bandwidth.explain_global| A global bandwidth
             is a single parameter that is shared amongst all kernels.
             While this is a fast and robust method,
             it is a rule of thumb approximation. Due to its global nature,
             it cannot take into account the different varying
             local densities. It uses notably the least amount of memory
             of all methods. |@docend:pdf.kde.bandwidth.explain_global|

    .. math::
      h = factor * min\left(\hat{\sigma}, \frac{IQR}{1.34}\right)\,n^{-\frac{1}{5}}

    Args:
        data: |@doc:pdf.kde.bandwidth.data| Data points to determine the bandwidth
               from. |@docend:pdf.kde.bandwidth.data|
        weights: |@doc:pdf.kde.bandwidth.weights| Weights of each event
               in *data*, can be None or Tensor-like with shape compatible
               with *data*. This will change the count of the events, whereas
               weight :math:`w_i` of :math:`x_i`. |@docend:pdf.kde.bandwidth.weights|
        factor (default: 0.9): Factor that scales the rule of thumb. Ofter used are 0.9 for
        silvermans rule of thumb or 1.059 for scotts rule of thumb.

    Returns:
        Estimated bandwidth
    """
    if factor is None:
        factor = tf.constant(0.9)
    return (
        min_std_or_iqr(data, weights)
        * tf.cast(tf.shape(data)[0], ztypes.float) ** (-1 / 5.0)
        * factor
    )


@z.function(wraps="tensor", keepalive=True)
def bandwidth_silverman(data, weights):
    r"""Calculate the bandwidth of *data* using silvermans rule of thumb.

    |@doc:pdf.kde.bandwidth.explain_global| A global bandwidth
             is a single parameter that is shared amongst all kernels.
             While this is a fast and robust method,
             it is a rule of thumb approximation. Due to its global nature,
             it cannot take into account the different varying
             local densities. It uses notably the least amount of memory
             of all methods. |@docend:pdf.kde.bandwidth.explain_global|

    .. math::
      h = 0.9 * min\left(\hat{\sigma}, \frac{IQR}{1.34}\right)\,n^{-\frac{1}{5}}

    Args:
        data: |@doc:pdf.kde.bandwidth.data| Data points to determine the bandwidth
               from. |@docend:pdf.kde.bandwidth.data|
        weights: |@doc:pdf.kde.bandwidth.weights| Weights of each event
               in *data*, can be None or Tensor-like with shape compatible
               with *data*. This will change the count of the events, whereas
               weight :math:`w_i` of :math:`x_i`. |@docend:pdf.kde.bandwidth.weights|

    Returns:
        Estimated bandwidth
    """
    return bandwidth_rule_of_thumb(
        data=data, weights=weights, factor=znp.array(0.9, dtype=ztypes.float)
    )


@z.function(wraps="tensor", keepalive=True)
def bandwidth_scott(data, weights):
    r"""Calculate the bandwidth of *data* using silvermans rule of thumb.

    |@doc:pdf.kde.bandwidth.explain_global| A global bandwidth
             is a single parameter that is shared amongst all kernels.
             While this is a fast and robust method,
             it is a rule of thumb approximation. Due to its global nature,
             it cannot take into account the different varying
             local densities. It uses notably the least amount of memory
             of all methods. |@docend:pdf.kde.bandwidth.explain_global|

    .. math::
      h = 1.059 * min\left(\hat{\sigma}, \frac{IQR}{1.34}\right)\,n^{-\frac{1}{5}}

    Args:
        data: |@doc:pdf.kde.bandwidth.data| Data points to determine the bandwidth
               from. |@docend:pdf.kde.bandwidth.data|
        weights: |@doc:pdf.kde.bandwidth.weights| Weights of each event
               in *data*, can be None or Tensor-like with shape compatible
               with *data*. This will change the count of the events, whereas
               weight :math:`w_i` of :math:`x_i`. |@docend:pdf.kde.bandwidth.weights|

    Returns:
        Estimated bandwidth
    """
    return bandwidth_rule_of_thumb(
        data=data, weights=weights, factor=znp.array(1.059, dtype=ztypes.float)
    )


def bandwidth_isj(data, weights):
    r"""Calculate the bandwidth of *data* using the improved Sheather-Jones Algorithm.

    The ISJ is an adaptive kernel density estimator based on linear diffusion processes
     and an estimation for the optimal bandwidth :footcite:t:`Botev_2010`

    |@doc:pdf.kde.bandwidth.explain_global| A global bandwidth
             is a single parameter that is shared amongst all kernels.
             While this is a fast and robust method,
             it is a rule of thumb approximation. Due to its global nature,
             it cannot take into account the different varying
             local densities. It uses notably the least amount of memory
             of all methods. |@docend:pdf.kde.bandwidth.explain_global|


    .. footbibliography::

    Args:
        data: |@doc:pdf.kde.bandwidth.data| Data points to determine the bandwidth
               from. |@docend:pdf.kde.bandwidth.data|
        weights: |@doc:pdf.kde.bandwidth.weights| Weights of each event
               in *data*, can be None or Tensor-like with shape compatible
               with *data*. This will change the count of the events, whereas
               weight :math:`w_i` of :math:`x_i`. |@docend:pdf.kde.bandwidth.weights|

    Returns:
        Estimated bandwidth
    """
    return isj_util.calculate_bandwidth(
        data, num_grid_points=1024, binning_method="linear", weights=weights
    )


def bandwidth_adaptive_geomV1(data, func, weights):
    r"""Local, adaptive bandwidth estimation scaling by the geometric mean.

    The implementation follows Eq. 2 and 3 of :footcite:t:`7761150`. However,
    experimental results hint towards a non-optimal performance, which can be caused
    by a mistake in the implementation.

    |@doc:pdf.kde.bandwidth.explain_adaptive| Adaptive bandwidths are
             a way to reduce the dependence on the bandwidth parameter
             and are usually local bandwidths that take into account
             the local densities.
             Adaptive bandwidths are constructed by using an initial estimate
             of the local density in order to calculate a sensible bandwidth
             for each kernel. The initial estimator can be a kernel density
             estimation using a global bandwidth with a rule of thumb.
             The adaptive bandwidth h is obtained using this estimate, where
             usually

             .. math::

               h_{i} \propto f( x_{i} ) ^ {-1/2}

             Estimates can still differ in the overall scaling of this
             bandwidth. |@docend:pdf.kde.bandwidth.explain_adaptive|

    |@doc:pdf.kde.bandwidth.explain_local| A local bandwidth
             means that each kernel :math:`i` has a different bandwidth.
             In other words, given some data points with size n,
             we will need n bandwidth parameters.
             This is often more accurate than a global bandwidth,
             as it allows to have larger bandwiths in areas of smaller density,
             where, due to the small local sample size, we have less certainty
             over the true density while having a smaller bandwidth in denser
             populated areas.

             The accuracy comes at the cost of a longer pre-calculation to obtain
             the local bandwidth (there are different methods available), an
             increased runtime and - most importantly - a peak memory usage.

             This can be especially costly for a large number (> few thousand) of
             kernels and/or evaluating on large datasets (> 10'000). |@docend:pdf.kde.bandwidth.explain_local|


    .. footbibliography::

    Args:
        data: |@doc:pdf.kde.bandwidth.data| Data points to determine the bandwidth
               from. |@docend:pdf.kde.bandwidth.data|
        weights: |@doc:pdf.kde.bandwidth.weights| Weights of each event
               in *data*, can be None or Tensor-like with shape compatible
               with *data*. This will change the count of the events, whereas
               weight :math:`w_i` of :math:`x_i`. |@docend:pdf.kde.bandwidth.weights|

    Returns:
        Estimated bandwidth of size data
    """
    data = z.convert_to_tensor(data)
    if weights is not None:
        n = znp.sum(weights)
    else:
        n = tf.cast(tf.shape(data)[0], ztypes.float)
    probs = func(data)
    lambda_i = 1 / znp.sqrt(
        probs / z.math.reduce_geometric_mean(probs, weights=weights)
    )

    return lambda_i * n ** (-1.0 / 5.0) * min_std_or_iqr(data, weights)


def bandwidth_adaptive_zfitV1(data, func, weights) -> znp.array:
    r"""(Naive) Local, adaptive bandwidth estimation using a normalized scaling.

    This approach is an ad-hoc scaling. It works well but is not found in any paper.

    |@doc:pdf.kde.bandwidth.explain_adaptive| Adaptive bandwidths are
             a way to reduce the dependence on the bandwidth parameter
             and are usually local bandwidths that take into account
             the local densities.
             Adaptive bandwidths are constructed by using an initial estimate
             of the local density in order to calculate a sensible bandwidth
             for each kernel. The initial estimator can be a kernel density
             estimation using a global bandwidth with a rule of thumb.
             The adaptive bandwidth h is obtained using this estimate, where
             usually

             .. math::

               h_{i} \propto f( x_{i} ) ^ {-1/2}

             Estimates can still differ in the overall scaling of this
             bandwidth. |@docend:pdf.kde.bandwidth.explain_adaptive|

    |@doc:pdf.kde.bandwidth.explain_local| A local bandwidth
             means that each kernel :math:`i` has a different bandwidth.
             In other words, given some data points with size n,
             we will need n bandwidth parameters.
             This is often more accurate than a global bandwidth,
             as it allows to have larger bandwiths in areas of smaller density,
             where, due to the small local sample size, we have less certainty
             over the true density while having a smaller bandwidth in denser
             populated areas.

             The accuracy comes at the cost of a longer pre-calculation to obtain
             the local bandwidth (there are different methods available), an
             increased runtime and - most importantly - a peak memory usage.

             This can be especially costly for a large number (> few thousand) of
             kernels and/or evaluating on large datasets (> 10'000). |@docend:pdf.kde.bandwidth.explain_local|


    .. footbibliography::

    Args:
        data: |@doc:pdf.kde.bandwidth.data| Data points to determine the bandwidth
               from. |@docend:pdf.kde.bandwidth.data|
        weights: |@doc:pdf.kde.bandwidth.weights| Weights of each event
               in *data*, can be None or Tensor-like with shape compatible
               with *data*. This will change the count of the events, whereas
               weight :math:`w_i` of :math:`x_i`. |@docend:pdf.kde.bandwidth.weights|

    Returns:
        Estimated bandwidth array of same size as data
    """
    data = z.convert_to_tensor(data)
    probs = func(data)
    estimate = bandwidth_scott(data, weights=weights)
    factor = znp.sqrt(probs) / znp.mean(znp.sqrt(probs))
    return estimate / factor


def bandwidth_adaptive_stdV1(data, func, weights):
    r"""Local, adaptive bandwidth estimation scaling by the std of the data.

    The implementation follows Eq. 2 and 3 of :footcite:t:`Cranmer_2001`. However,
    experimental results hint towards a non-optimal performance, which can be caused
    by a mistake in the implementation.

    |@doc:pdf.kde.bandwidth.explain_adaptive| Adaptive bandwidths are
             a way to reduce the dependence on the bandwidth parameter
             and are usually local bandwidths that take into account
             the local densities.
             Adaptive bandwidths are constructed by using an initial estimate
             of the local density in order to calculate a sensible bandwidth
             for each kernel. The initial estimator can be a kernel density
             estimation using a global bandwidth with a rule of thumb.
             The adaptive bandwidth h is obtained using this estimate, where
             usually

             .. math::

               h_{i} \propto f( x_{i} ) ^ {-1/2}

             Estimates can still differ in the overall scaling of this
             bandwidth. |@docend:pdf.kde.bandwidth.explain_adaptive|

    |@doc:pdf.kde.bandwidth.explain_local| A local bandwidth
             means that each kernel :math:`i` has a different bandwidth.
             In other words, given some data points with size n,
             we will need n bandwidth parameters.
             This is often more accurate than a global bandwidth,
             as it allows to have larger bandwiths in areas of smaller density,
             where, due to the small local sample size, we have less certainty
             over the true density while having a smaller bandwidth in denser
             populated areas.

             The accuracy comes at the cost of a longer pre-calculation to obtain
             the local bandwidth (there are different methods available), an
             increased runtime and - most importantly - a peak memory usage.

             This can be especially costly for a large number (> few thousand) of
             kernels and/or evaluating on large datasets (> 10'000). |@docend:pdf.kde.bandwidth.explain_local|


    .. footbibliography::

    Args:
        data: |@doc:pdf.kde.bandwidth.data| Data points to determine the bandwidth
               from. |@docend:pdf.kde.bandwidth.data|
        weights: |@doc:pdf.kde.bandwidth.weights| Weights of each event
               in *data*, can be None or Tensor-like with shape compatible
               with *data*. This will change the count of the events, whereas
               weight :math:`w_i` of :math:`x_i`. |@docend:pdf.kde.bandwidth.weights|

    Returns:
        Estimated bandwidth array of same size as data
    """
    data = z.convert_to_tensor(data)
    if weights is not None:
        n = znp.sum(weights)
    else:
        n = tf.cast(tf.shape(data)[0], ztypes.float)
    probs = func(data)
    divisor = min_std_or_iqr(data, weights)
    bandwidth = z.sqrt(divisor / probs)
    bandwidth *= tf.cast(n, ztypes.float) ** (-1.0 / 5.0) * 1.059
    return bandwidth


def adaptive_factory(func, grid):
    if grid:

        def adaptive(constructor, data, **kwargs):
            kwargs.pop("name", None)
            kde_silverman = constructor(
                bandwidth="silverman",
                data=data,
                name="INTERNAL_for_adaptive_kde",
                **kwargs,
            )
            grid = kde_silverman._grid
            weights = kde_silverman._grid_data
            return func(
                data=grid,
                func=kde_silverman.pdf,
                weights=weights * tf.cast(tf.shape(data)[0], ztypes.float),
            )

    else:

        def adaptive(constructor, data, weights, **kwargs):
            kwargs.pop("name", None)
            kde_silverman = constructor(
                bandwidth="silverman",
                data=data,
                name="INTERNAL_for_adaptive_kde",
                **kwargs,
            )
            return func(data=data, func=kde_silverman.pdf, weights=weights)

    return adaptive


_adaptive_geom_bandwidth_grid_KDEV1 = adaptive_factory(
    bandwidth_adaptive_geomV1, grid=True
)
_adaptive_geom_bandwidth_KDEV1 = adaptive_factory(bandwidth_adaptive_geomV1, grid=False)

_adaptive_std_bandwidth_grid_KDEV1 = adaptive_factory(
    bandwidth_adaptive_stdV1, grid=True
)
_adaptive_std_bandwidth_KDEV1 = adaptive_factory(bandwidth_adaptive_stdV1, grid=False)

_adaptive_zfit_bandwidth_grid_KDEV1 = adaptive_factory(
    bandwidth_adaptive_zfitV1, grid=True
)
_adaptive_zfit_bandwidth_KDEV1 = adaptive_factory(bandwidth_adaptive_zfitV1, grid=False)


def _bandwidth_scott_KDEV1(data, weights, *_, **__):
    return bandwidth_scott(
        data,
        weights=weights,
    )


def _bandwidth_silverman_KDEV1(data, weights, *_, **__):
    return bandwidth_silverman(
        data,
        weights=weights,
    )


def _bandwidth_isj_KDEV1(data, weights, *_, **__):
    return bandwidth_isj(data, weights=weights)


def check_bw_grid_shapes(bandwidth, grid=None, n_grid=None):
    if run.executing_eagerly() and bw_is_arraylike(bandwidth, allow1d=False):
        n_grid = grid.shape[0] if grid is not None else n_grid
        if n_grid is None:
            raise ValueError("Either the grid or n_grid must be given.")
        if bandwidth.shape[0] != n_grid:
            raise ShapeIncompatibleError(
                "The bandwidth array must have the same length as the grid"
            )


@z.function(wraps="tensor", keepalive=True)
def min_std_or_iqr(x, weights):
    if weights is not None:
        return znp.minimum(
            znp.sqrt(tf.nn.weighted_moments(x, axes=[0], frequency_weights=weights)[1]),
            weighted_quantile(x, 0.75, weights=weights)[0]
            - weighted_quantile(x, 0.25, weights=weights)[0],
        )
    else:
        return znp.minimum(
            znp.std(x), (tfp.stats.percentile(x, 75) - tfp.stats.percentile(x, 25))
        )


@z.function(wraps="tensor", keepalive=True)
def calc_kernel_probs(size, weights):
    if weights is not None:
        return weights / znp.sum(weights)
    else:
        return tf.broadcast_to(1 / size, shape=(tf.cast(size, tf.int32),))


class KDEHelper:
    _bandwidth_methods = {
        "scott": _bandwidth_scott_KDEV1,
        "silverman": _bandwidth_silverman_KDEV1,
    }
    _default_padding = False
    _default_num_grid_points = 1024

    def _convert_init_data_weights_size(
        self, data, weights, padding, limits=None, bandwidth=None
    ):
        self._original_data = data  # for copying
        if isinstance(data, ZfitData):
            if data.weights is not None:
                if weights is not None:
                    raise OverdefinedError(
                        "Cannot specify weights and use a `ZfitData` with weights."
                    )
                else:
                    weights = data.weights

            if data.n_obs > 1:
                raise ShapeIncompatibleError(
                    f"KDE is 1 dimensional, but data {data} has {data.n_obs} observables."
                )
            data = z.unstack_x(data)

        if callable(padding):
            data, weights, bandwidth = padding(
                data=data, weights=weights, limits=limits, bandwidth=bandwidth
            )
        elif padding is not False:
            data, weights, bandwidth = padreflect_data_weights_1dim(
                data, weights=weights, mode=padding, limits=limits, bandwidth=bandwidth
            )
        shape_data = tf.shape(data)
        size = tf.cast(shape_data[0], ztypes.float)
        return data, size, weights, bandwidth

    def _convert_input_bandwidth(self, bandwidth, data, **kwargs):
        if bandwidth is None:
            bandwidth = "silverman"
        # estimate bandwidth
        bandwidth_param = bandwidth
        if isinstance(bandwidth, str):
            bandwidth = self._bandwidth_methods.get(bandwidth)
            if bandwidth is None:
                raise ValueError(
                    f"Cannot use {bandwidth} as a bandwidth method. Use a numerical value or one of"
                    f" the defined methods: {list(self._bandwidth_methods.keys())}"
                )
        if (not isinstance(bandwidth, ZfitParameter)) and callable(bandwidth):
            bandwidth = bandwidth(constructor=type(self), data=data, **kwargs)
        is_arraylike = bw_is_arraylike(bandwidth_param, allow1d=True)
        if (
            bandwidth_param is None
            or is_arraylike
            or bandwidth_param
            in (
                "adaptiveV1",
                "adaptive",
                "adaptive_zfit",
                "adaptive_std",
                "adaptive_geom",
            )
        ):
            bandwidth_param = -999999  # dummy value, not needed. Better solution?
        else:
            bandwidth_param = bandwidth
        if is_arraylike:
            bandwidth = znp.asarray(bandwidth)
        return bandwidth, bandwidth_param


def padreflect_data_weights_1dim(data, mode, weights=None, limits=None, bandwidth=None):
    if mode is False:
        return data, weights
    if mode is True:
        mode = znp.array(0.1)
    if not isinstance(mode, dict):
        mode = {"lowermirror": mode, "uppermirror": mode}
    for key in mode:
        if key not in ("lowermirror", "uppermirror"):
            raise ValueError(
                f"Key '{key}' is not a valid padding specification, use 'lowermirror' or 'uppermirror'"
                f" in order to mirror the data."
            )
    if limits is None:
        minimum = znp.min(data)
        maximum = znp.max(data)
    else:
        minimum = znp.array(limits[0][0])
        maximum = znp.array(limits[1][0])

    diff = maximum - minimum
    new_data = []
    new_weights = []
    bw_is_array = bw_is_arraylike(bandwidth, allow1d=False)
    if bw_is_array:
        new_bw = []

    if (lower := mode.get("lowermirror")) is not None:
        dx_lower = diff * lower
        lower_area = data < minimum + dx_lower
        lower_index = znp.where(lower_area)[0]
        lower_data = tf.gather(data, indices=lower_index)
        lower_data_mirrored = -lower_data + 2 * minimum
        new_data.append(lower_data_mirrored)
        if weights is not None:
            lower_weights = tf.gather(weights, indices=lower_index)
            new_weights.append(lower_weights)
        if bw_is_array:
            new_bw.append(tf.gather(bandwidth, indices=lower_index))
    new_data.append(data)
    new_weights.append(weights)
    if bw_is_array:
        new_bw.append(bandwidth)
    if (upper := mode.get("uppermirror")) is not None:
        dx_upper = diff * upper
        upper_area = data > maximum - dx_upper
        upper_index = znp.where(upper_area)[0]
        upper_data = tf.gather(data, indices=upper_index)
        upper_data_mirrored = -upper_data + 2 * maximum
        new_data.append(upper_data_mirrored)
        if weights is not None:
            upper_weights = tf.gather(weights, indices=upper_index)
            new_weights.append(upper_weights)
        if bw_is_array:
            new_bw.append(tf.gather(bandwidth, indices=upper_index))
    data = tf.concat(new_data, axis=0)

    if weights is not None:
        weights = tf.concat(new_weights, axis=0)
    if bw_is_array:
        bandwidth = tf.concat(new_bw, axis=0)
    return data, weights, bandwidth


class GaussianKDE1DimV1(KDEHelper, WrapDistribution):
    _N_OBS = 1
    _bandwidth_methods = KDEHelper._bandwidth_methods.copy()
    _bandwidth_methods.update(
        {"adaptive": _adaptive_std_bandwidth_KDEV1, "isj": _bandwidth_isj_KDEV1}
    )

    def __init__(
        self,
        obs: ztyping.ObsTypeInput,
        data: ztyping.ParamTypeInput,
        bandwidth: ztyping.ParamTypeInput | str = None,
        weights: None | np.ndarray | tf.Tensor = None,
        truncate: bool = False,
        *,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name: str = "GaussianKDE1DimV1",
    ):
        r"""EXPERIMENTAL, `FEEDBACK WELCOME.

        <https://github.com/zfit/zfit/issues/new?assignees=&labels=&template=other.md&title=>`_ Exact, one dimensional,
        (truncated) Kernel Density Estimation with a Gaussian Kernel.

        This implementation features an exact implementation as is preferably used for smaller (max. ~ a few thousand
        points) data sets. For larger data sets, methods such as :py:class:`~zfit.pdf.KDE1DimGrid`
        that bin the dataset may be more efficient
        Kernel Density Estimation is a non-parametric method to approximate the density of given points.

        .. math::

            f_h(x) =  \frac{1}{nh} \sum_{i=1}^n K\Big(\frac{x-x_i}{h}\Big)

        where the kernel in this case is a (truncated) Gaussian

        .. math::
            K = \exp \Big(\frac{(x - x_i)^2}{\sigma^2}\Big)


        The bandwidth of the kernel can be estimated in different ways. It can either be a global bandwidth,
        corresponding to a single value, or a local bandwidth, each corresponding to one data point.

        **Usage**

        The KDE can be instantiated by using a numpy-like data sample, preferably a `zfit.Data` object. This
        will be used as the mean of the kernels. The bandwidth can either be given as a parameter (with length
        1 for a global one or length equal to the data size) - a rather advanced concept for methods such as
        cross validation - or determined from the data automatically, either through a simple method like
        scott or silverman rule of thumbs or through an iterative, adaptive method.

        Examples
        --------

        .. code-block:: python

            # generate some example kernels
            size = 150
            data = np.random.normal(size=size, loc=2, scale=3)

            limits = (-15, 5)
            obs = zfit.Space("obs1", limits=limits)
            kde_silverman = zfit.pdf.GaussianKDE1DimV1(data=data, obs=obs, bandwidth='silverman')

            # for a better smoothing of the kernels, use an adaptive approach
            kde = zfit.pdf.GaussianKDE1DimV1(data=data, obs=obs, bandwidth='adaptive')


        Args:
            data: |@doc:pdf.kde.init.data| Data sample to approximate
             the density from. The points represent positions of the *kernel*,
             the :math:`x_i`. This is preferrably a ``ZfitData``, but can also
             be an array-like object.

             If the data has weights, they will be taken into account.
             This will change the count of the events, whereas
             weight :math:`w_i` of :math:`x_i` will scale the value of
             :math:`K_i( x_i)`, resulting in a factor of :math:`\frac{w_i}{\sum w_i} `.

             If no weights are given, each kernel will be scaled by the same
             constant :math:`\frac{1}{n_{data}}`. |@docend:pdf.kde.init.data|

            obs: |@doc:pdf.kde.init.obs| Observable space of the KDE.
             As with any other PDF, this will be used as the default *norm*, but
             does not define the domain of the PDF. Namely, this can be a smaller
             space than *data*, as long as the name of the observable match.
             Using a larger dataset is actually good practice avoiding
             bountary biases, see also :ref:`sec-boundary-bias-and-padding`. |@docend:pdf.kde.init.obs|
            bandwidth: Valid pre-defined options are {'silverman', 'scott', 'adaptive'}.
             |@doc:pdf.kde.init.bandwidth| Bandwidth of the kernel,
             often also denoted as :math:`h`. For a Gaussian kernel, this
             corresponds to *sigma*. This can be calculated using
             pre-defined options or by specifying a numerical value that is
             broadcastable to *data* -- a scalar or an array-like
             object with the same size as *data*.

             A scalar value is usually referred to as a global bandwidth while
             an array holds local bandwidths |@docend:pdf.kde.init.bandwidth|

            The bandwidth can also be a parameter, which should be used with caution. However,
            it allows to use it in cross-valitadion with a likelihood method.


            weights: |@doc:pdf.kde.init.weights| Weights of each event
             in *data*, can be None or Tensor-like with shape compatible
             with *data*. Instead of using this parameter, it is preferred
             to use a ``ZfitData`` as *data* that contains weights.
             This will change the count of the events, whereas
             weight :math:`w_i` of :math:`x_i` will scale the value of :math:`K_i( x_i)`,
             resulting in a factor of :math:`\frac{w_i}{\sum w_i} `.

             If no weights are given, each kernel will be scaled by the same
             constant :math:`\frac{1}{n_{data}}`. |@docend:pdf.kde.init.weights|
            truncate: If a truncated Gaussian kernel should be used with the limits given by the `obs` lower and
                upper limits. This can cause NaNs in case datapoints are outside the limits.
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: |@doc:pdf.init.name| Human-readable name
               or label of
               the PDF for better identification.
               Has no programmatical functional purpose as identification. |@docend:pdf.init.name|
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
        """
        original_data = data
        data, size, weights, _ = self._convert_init_data_weights_size(
            data,
            weights,
            padding=False,
            limits=None,
            bandwidth=bandwidth,
        )

        bandwidth, bandwidth_param = self._convert_input_bandwidth(
            bandwidth=bandwidth,
            data=data,
            truncate=truncate,
            name=name,
            obs=obs,
            weights=weights,
        )
        params = {"bandwidth": bandwidth_param}

        probs = calc_kernel_probs(size, weights)
        categorical = tfd.Categorical(probs=probs)  # no grad -> no need to recreate

        # create distribution factory
        if truncate:
            if not isinstance(obs, ZfitSpace):
                raise ValueError(
                    "`obs` has to be a `ZfitSpace` if `truncated` is True."
                )
            inside = obs.inside(data)
            all_inside = znp.all(inside)
            tf.debugging.assert_equal(
                all_inside,
                True,
                message="Not all data points are inside the limits but"
                " a truncate kernel was chosen.",
            )

            def kernel_factory():
                return tfp.distributions.TruncatedNormal(
                    loc=self._data,
                    scale=self._bandwidth,
                    low=self.space.rect_lower,
                    high=self.space.rect_upper,
                )

        else:

            def kernel_factory():
                return tfp.distributions.Normal(loc=self._data, scale=self._bandwidth)

        def dist_kwargs():
            return dict(
                mixture_distribution=categorical,
                components_distribution=kernel_factory(),
            )

        distribution = tfd.MixtureSameFamily

        super().__init__(
            obs=obs,
            params=params,
            dist_params={},
            dist_kwargs=dist_kwargs,
            distribution=distribution,
            extended=extended,
            norm=norm,
            name=name,
        )

        self._data_weights = weights
        self._bandwidth = bandwidth
        self._data = data
        self._original_data = original_data  # for copying
        self._truncate = truncate


class KDE1DimExact(KDEHelper, WrapDistribution, SerializableMixin):
    _bandwidth_methods = KDEHelper._bandwidth_methods.copy()
    _bandwidth_methods.update(
        {
            "adaptive_geom": _adaptive_geom_bandwidth_KDEV1,
            "adaptive_std": _adaptive_std_bandwidth_KDEV1,
            "adaptive_zfit": _adaptive_zfit_bandwidth_KDEV1,
            "isj": _bandwidth_isj_KDEV1,
        }
    )

    def __init__(
        self,
        data: ztyping.XTypeInput,
        *,
        obs: ztyping.ObsTypeInput | None = None,
        bandwidth: ztyping.ParamTypeInput | str | Callable | None = None,
        kernel: tfd.Distribution = None,
        padding: callable | str | bool | None = None,
        weights: np.ndarray | tf.Tensor | None = None,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name: str | None = "ExactKDE1DimV1",
    ):
        r"""Kernel Density Estimation is a non-parametric method to approximate the density of given points.

        For a more in-depth explanation, see also in the section about Kernel Density Estimation
        :ref:`section-exact-kdes`

        Given a sample *data* we want to estimate the

        .. math::
            f_h(x) =  \frac{1}{nh} \sum_{i=1}^n K\Big(\frac{x-x_i}{h}\Big)

        This PDF features an exact implementation as is preferable for smaller (max. ~ a few thousand
        points) data sets. For larger data sets, methods such as :py:class:`~zfit.pdf.KDE1DimGrid`
        that bin the dataset may be more efficient
        Kernel Density Estimation is a non-parametric method to approximate the density of given points.

        .. math::

            f_h(x) =  \frac{1}{nh} \sum_{i=1}^n K\Big(\frac{x-x_i}{h}\Big)

        where the kernel in this case is a (truncated) Gaussian

        .. math::
            K = \exp \Big(\frac{(x - x_i)^2}{\sigma^2}\Big)


        The bandwidth of the kernel can be estimated in different ways. It can either be a global bandwidth,
        corresponding to a single value, or a local bandwidth, each corresponding to one data point.

        Args:
            data: |@doc:pdf.kde.init.data| Data sample to approximate
             the density from. The points represent positions of the *kernel*,
             the :math:`x_i`. This is preferrably a ``ZfitData``, but can also
             be an array-like object.

             If the data has weights, they will be taken into account.
             This will change the count of the events, whereas
             weight :math:`w_i` of :math:`x_i` will scale the value of
             :math:`K_i( x_i)`, resulting in a factor of :math:`\frac{w_i}{\sum w_i} `.

             If no weights are given, each kernel will be scaled by the same
             constant :math:`\frac{1}{n_{data}}`. |@docend:pdf.kde.init.data|

            obs: |@doc:pdf.kde.init.obs| Observable space of the KDE.
             As with any other PDF, this will be used as the default *norm*, but
             does not define the domain of the PDF. Namely, this can be a smaller
             space than *data*, as long as the name of the observable match.
             Using a larger dataset is actually good practice avoiding
             bountary biases, see also :ref:`sec-boundary-bias-and-padding`. |@docend:pdf.kde.init.obs|
            bandwidth: Valid pre-defined options are {'silverman', 'scott',
             'adaptive_zfit', 'adaptive_geom', 'adaptive_std', 'isj'}.
             |@doc:pdf.kde.init.bandwidth| Bandwidth of the kernel,
             often also denoted as :math:`h`. For a Gaussian kernel, this
             corresponds to *sigma*. This can be calculated using
             pre-defined options or by specifying a numerical value that is
             broadcastable to *data* -- a scalar or an array-like
             object with the same size as *data*.

             A scalar value is usually referred to as a global bandwidth while
             an array holds local bandwidths |@docend:pdf.kde.init.bandwidth|
             The bandwidth can also be a parameter, which should be used with caution. However,
             it allows to use it in cross-valitadion with a likelihood method.

            kernel: |@doc:pdf.kde.init.kernel| The kernel is the heart
             of the Kernel Density Estimation, which consists of the sum of
             kernels around each sample point. Therefore, a kernel should represent
             the distribution probability of a single data point as close as
             possible.

             The most widespread kernel is a Gaussian, or Normal, distribution. Due
             to the law of large numbers, the sum of many (arbitrary) random variables
             -- this is the case for most real world observable as they are the result of
             multiple consecutive random effects -- results in a Gaussian distribution.
             However, there are many cases where this assumption is not per-se true. In
             this cases an alternative kernel may offer a better choice.

             Valid choices are callables that return a
             :py:class:`~tensorflow_probability.distribution.Distribution`, such as all distributions
             that belong to the loc-scale family. |@docend:pdf.kde.init.kernel|

            padding: |@doc:pdf.kde.init.padding| KDEs have a peculiar
             weakness: the boundaries, as the outside has a zero density. This makes the KDE
             go down at the bountary as well, as the density approaches zero, no matter what the
             density inside the boundary was.

             There are two ways to circumvent this problem:

               - the best solution: providing a larger dataset than the default space the PDF is used in
               - mirroring the existing data at the boundaries, which is equivalent to a boundary condition
                 with a zero derivative. This is a padding technique and can improve the boundaries.
                 However, one important drawback of this method is to keep in mind that this will actually
                 alter the PDF *to look mirrored*. If the PDF is plotted in a larger range, this becomes
                 clear.

             Possible options are a number (default 0.1) that depicts the fraction of the overall space
             that defines the data mirrored on both sides. For example, for a space from 0 to 5, a value of
             0.3 means that all data in the region of 0 to 1.5 is taken, mirrored around 0 as well as
             all data from 3.5 to 5 and mirrored at 5. The new data will go from -1.5 to 6.5, so the
             KDE is also having a shape outside the desired range. Using it only for the range 0 to 5
             hides this.
             Using a dict, each side separately (or only a single one) can be mirrored, like ``{'lowermirror: 0.1}``
             or ``{'lowermirror: 0.2, 'uppermirror': 0.1}``.
             For more control, a callable that takes data and weights can also be used. |@docend:pdf.kde.init.padding|

            weights: |@doc:pdf.kde.init.weights| Weights of each event
             in *data*, can be None or Tensor-like with shape compatible
             with *data*. Instead of using this parameter, it is preferred
             to use a ``ZfitData`` as *data* that contains weights.
             This will change the count of the events, whereas
             weight :math:`w_i` of :math:`x_i` will scale the value of :math:`K_i( x_i)`,
             resulting in a factor of :math:`\frac{w_i}{\sum w_i} `.

             If no weights are given, each kernel will be scaled by the same
             constant :math:`\frac{1}{n_{data}}`. |@docend:pdf.kde.init.weights|
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: |@doc:model.init.name| Human-readable name
               or label of
               the PDF for better identification.
               Has no programmatical functional purpose as identification. |@docend:model.init.name|
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
        """
        original_init = {
            "data": data,
            "obs": obs,
            "bandwidth": bandwidth,
            "kernel": kernel,
            "padding": padding,
            "weights": weights,
            "extended": extended,
            "name": name,
        }

        if kernel is None:
            kernel = tfd.Normal

        if padding is None:
            padding = self._default_padding
        if obs is None:
            if not isinstance(data, ZfitData) or not data.space.has_limits:
                raise ValueError(
                    "obs can only be None if data is ZfitData with limits."
                )
            else:
                obs = data.space
        data, size, weights, bandwidth = self._convert_init_data_weights_size(
            data, weights, padding=padding, limits=obs.limits, bandwidth=bandwidth
        )
        self._padding = padding
        bandwidth, bandwidth_param = self._convert_input_bandwidth(
            bandwidth=bandwidth,
            data=data,
            name=name,
            obs=obs,
            weights=weights,
            padding=False,
            kernel=kernel,
        )

        self._original_data = data  # for copying

        def components_distribution_generator(loc, scale):
            return tfd.Independent(kernel(loc=loc, scale=scale))

        self._data = data
        self._bandwidth = bandwidth
        self._kernel = kernel
        self._weights = weights

        probs = calc_kernel_probs(size, weights)

        mixture_distribution = tfd.Categorical(probs=probs)
        components_distribution = components_distribution_generator(
            loc=self._data, scale=self._bandwidth
        )

        def dist_kwargs():
            return dict(
                mixture_distribution=mixture_distribution,
                components_distribution=components_distribution,
            )

        distribution = tfd.MixtureSameFamily

        params = {"bandwidth": bandwidth_param}

        super().__init__(
            obs=obs,
            params=params,
            dist_params={},
            dist_kwargs=dist_kwargs,
            distribution=distribution,
            extended=extended,
            norm=norm,
            name=name,
        )
        self.hs3.original_init.update(original_init)


class KDE1DimExactRepr(BasePDFRepr):
    _implementation = KDE1DimExact
    hs3_type: Literal["KDE1DimExact"] = pydantic.Field("KDE1DimExact", alias="type")

    data: Union[np.ndarray, Serializer.types.DataTypeDiscriminated]
    obs: Optional[SpaceRepr] = None
    bandwidth: Optional[Union[str, float]] = None
    kernel: None = None
    padding: Optional[Union[bool, str]] = None
    weights: Optional[Union[np.ndarray, tf.Tensor]] = None
    name: Optional[str] = "KDE1DimExact"

    @pydantic.validator("kernel", pre=True)
    def validate_kernel(cls, v):
        if v is not None:
            if v != tfd.Normal:
                raise ValueError(
                    "Kernel must be None for KDE1DimExact to be serialized."
                )
            else:
                v = None
        return v

    @pydantic.root_validator(pre=True)
    def validate_all(cls, values):
        values = dict(values)
        if cls.orm_mode(values):
            for k, v in values["hs3"].original_init.items():
                values[k] = v
        return values


class KDE1DimGrid(KDEHelper, WrapDistribution, SerializableMixin):
    _N_OBS = 1
    _bandwidth_methods = KDEHelper._bandwidth_methods.copy()
    _bandwidth_methods.update(
        {
            "adaptive_geom": _adaptive_geom_bandwidth_grid_KDEV1,
            "adaptive_zfit": _adaptive_zfit_bandwidth_grid_KDEV1,
            # 'adaptive_std': _adaptive_std_bandwidth_grid_KDEV1,
        }
    )

    def __init__(
        self,
        data: ztyping.XTypeInput,
        *,
        bandwidth: ztyping.ParamTypeInput | str | Callable | None = None,
        kernel: tfd.Distribution = None,
        padding: callable | str | bool | None = None,
        num_grid_points: int | None = None,
        binning_method: str | None = None,
        obs: ztyping.ObsTypeInput | None = None,
        weights: np.ndarray | tf.Tensor | None = None,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name: str = "GridKDE1DimV1",
    ):
        r"""Kernel Density Estimation is a non-parametric method to approximate the density of given points.

        For a more in-depth explanation, see also in the section about Kernel Density Estimation
        :ref:`sec-grid-kdes`

        .. math::
            f_h(x) =  \frac{1}{nh} \sum_{i=1}^n K\Big(\frac{x-x_i}{h}\Big)

        Args:
            data: |@doc:pdf.kde.init.data| Data sample to approximate
             the density from. The points represent positions of the *kernel*,
             the :math:`x_i`. This is preferrably a ``ZfitData``, but can also
             be an array-like object.

             If the data has weights, they will be taken into account.
             This will change the count of the events, whereas
             weight :math:`w_i` of :math:`x_i` will scale the value of
             :math:`K_i( x_i)`, resulting in a factor of :math:`\frac{w_i}{\sum w_i} `.

             If no weights are given, each kernel will be scaled by the same
             constant :math:`\frac{1}{n_{data}}`. |@docend:pdf.kde.init.data|

            bandwidth: Valid pre-defined options are {'silverman', 'scott',
             'adaptive_zfit', 'adaptive_geom'}.
             |@doc:pdf.kde.init.bandwidth| Bandwidth of the kernel,
             often also denoted as :math:`h`. For a Gaussian kernel, this
             corresponds to *sigma*. This can be calculated using
             pre-defined options or by specifying a numerical value that is
             broadcastable to *data* -- a scalar or an array-like
             object with the same size as *data*.

             A scalar value is usually referred to as a global bandwidth while
             an array holds local bandwidths |@docend:pdf.kde.init.bandwidth|
            kernel: |@doc:pdf.kde.init.kernel| The kernel is the heart
             of the Kernel Density Estimation, which consists of the sum of
             kernels around each sample point. Therefore, a kernel should represent
             the distribution probability of a single data point as close as
             possible.

             The most widespread kernel is a Gaussian, or Normal, distribution. Due
             to the law of large numbers, the sum of many (arbitrary) random variables
             -- this is the case for most real world observable as they are the result of
             multiple consecutive random effects -- results in a Gaussian distribution.
             However, there are many cases where this assumption is not per-se true. In
             this cases an alternative kernel may offer a better choice.

             Valid choices are callables that return a
             :py:class:`~tensorflow_probability.distribution.Distribution`, such as all distributions
             that belong to the loc-scale family. |@docend:pdf.kde.init.kernel|
            padding: |@doc:pdf.kde.init.padding| KDEs have a peculiar
             weakness: the boundaries, as the outside has a zero density. This makes the KDE
             go down at the bountary as well, as the density approaches zero, no matter what the
             density inside the boundary was.

             There are two ways to circumvent this problem:

               - the best solution: providing a larger dataset than the default space the PDF is used in
               - mirroring the existing data at the boundaries, which is equivalent to a boundary condition
                 with a zero derivative. This is a padding technique and can improve the boundaries.
                 However, one important drawback of this method is to keep in mind that this will actually
                 alter the PDF *to look mirrored*. If the PDF is plotted in a larger range, this becomes
                 clear.

             Possible options are a number (default 0.1) that depicts the fraction of the overall space
             that defines the data mirrored on both sides. For example, for a space from 0 to 5, a value of
             0.3 means that all data in the region of 0 to 1.5 is taken, mirrored around 0 as well as
             all data from 3.5 to 5 and mirrored at 5. The new data will go from -1.5 to 6.5, so the
             KDE is also having a shape outside the desired range. Using it only for the range 0 to 5
             hides this.
             Using a dict, each side separately (or only a single one) can be mirrored, like ``{'lowermirror: 0.1}``
             or ``{'lowermirror: 0.2, 'uppermirror': 0.1}``.
             For more control, a callable that takes data and weights can also be used. |@docend:pdf.kde.init.padding|
            num_grid_points: |@doc:pdf.kde.init.num_grid_points| Number of points in
             the binning grid.

             The data will be binned using the *binning_method* in *num_grid_points*
             and this histogram grid will then be used as kernel points. This has the
             advantage to have a constant computational complexity independent of the data
             size.

             A number from 32 on can already yield good results, while the default is set
             to 1024, creating a fine grid. Lowering the number increases the performance
             at the cost of accuracy. |@docend:pdf.kde.init.num_grid_points|
            binning_method: |@doc:pdf.kde.init.binning_method| Method to be used for
             binning the data. Options are 'linear', 'simple'.

             The data can be binned in the usual way ('simple'), but this is less precise
             for KDEs, where we are interested in the shape of the histogram and smoothing
             it. Therefore, a better suited method, 'linear', is available.

             In normal binnig, each event (or weight) falls into the bin within the bin edges,
             while the neighbouring bins get zero counts from this event.
             In linear binning, the event is split between two bins, proportional to its
             closeness to each bin.

             The 'linear' method provides superior performance, most notably in small (~32)
             grids. |@docend:pdf.kde.init.binning_method|
            obs: |@doc:pdf.kde.init.obs| Observable space of the KDE.
             As with any other PDF, this will be used as the default *norm*, but
             does not define the domain of the PDF. Namely, this can be a smaller
             space than *data*, as long as the name of the observable match.
             Using a larger dataset is actually good practice avoiding
             bountary biases, see also :ref:`sec-boundary-bias-and-padding`. |@docend:pdf.kde.init.obs|
            weights: |@doc:pdf.kde.init.weights| Weights of each event
             in *data*, can be None or Tensor-like with shape compatible
             with *data*. Instead of using this parameter, it is preferred
             to use a ``ZfitData`` as *data* that contains weights.
             This will change the count of the events, whereas
             weight :math:`w_i` of :math:`x_i` will scale the value of :math:`K_i( x_i)`,
             resulting in a factor of :math:`\frac{w_i}{\sum w_i} `.

             If no weights are given, each kernel will be scaled by the same
             constant :math:`\frac{1}{n_{data}}`. |@docend:pdf.kde.init.weights|
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: |@doc:model.init.name| Human-readable name
               or label of
               the PDF for better identification.
               Has no programmatical functional purpose as identification. |@docend:model.init.name|
            extended: |@doc:model.init.extended| Whether the PDF is extended
                or not. If True, the PDF can be integrated over the full space
                and the integral will be 1. If False, the integral will be the
                number of events in the dataset. |@docend:model.init.extended|
        """
        original_init = {
            "data": data,
            "bandwidth": bandwidth,
            "kernel": kernel,
            "binning_method": binning_method,
            "num_grid_points": num_grid_points,
            "obs": obs,
            "weights": weights,
            "padding": padding,
            "name": name,
            "extended": extended,
            "norm": norm,
        }
        if kernel is None:
            kernel = tfd.Normal
        if binning_method is None:
            binning_method = "linear"
        if num_grid_points is None:
            num_grid_points = self._default_num_grid_points

        if isinstance(
            bandwidth, str
        ):  # numpy arrays cannot be compared with equal-> "use any, all"
            if bandwidth == "isj":
                raise ValueError(
                    "isj not supported in GridKDE, use directly 'KDE1DimISJ'"
                )
            if bandwidth == "adaptive_std":
                raise ValueError(
                    "adaptive_std not supported in GridKDE due to very bad results. This is maybe caused"
                    " by an issue regarding weights of the underlaying implementation."
                )

        if padding is None:
            padding = self._default_padding
        if obs is None:
            if not isinstance(data, ZfitData) or not data.space.has_limits:
                raise ValueError(
                    "obs can only be None if data is ZfitData with limits."
                )
            else:
                obs = data.space
        data, size, weights, _ = self._convert_init_data_weights_size(
            data, weights, padding=padding, limits=obs.limits, bandwidth=bandwidth
        )
        self._padding = padding

        self._original_data = data  # for copying

        def components_distribution_generator(loc, scale):
            return tfd.Independent(kernel(loc=loc, scale=scale))

        if num_grid_points is not None:
            num_grid_points = tf.minimum(
                tf.cast(size, ztypes.int), tf.cast(num_grid_points, ztypes.int)
            )
        self._num_grid_points = num_grid_points
        self._binning_method = binning_method
        self._data = data
        self._grid = binning_util.generate_1d_grid(
            self._data, num_grid_points=self._num_grid_points
        )

        bandwidth, bandwidth_param = self._convert_input_bandwidth(
            bandwidth=bandwidth,
            data=data,
            binning_method=binning_method,
            num_grid_points=num_grid_points,
            padding=False,
            kernel=kernel,
            name=name,
            obs=obs,
            weights=weights,
        )

        self._bandwidth = bandwidth
        self._kernel = kernel
        self._weights = weights

        self._grid_data = binning_util.bin_1d(
            self._binning_method, self._data, self._grid, self._weights
        )

        mixture_distribution = tfd.Categorical(probs=self._grid_data)

        check_bw_grid_shapes(self._bandwidth, self._grid)

        components_distribution = components_distribution_generator(
            loc=self._grid, scale=self._bandwidth
        )

        def dist_kwargs():
            return dict(
                mixture_distribution=mixture_distribution,
                components_distribution=components_distribution,
            )

        distribution = tfd.MixtureSameFamily

        params = {"bandwidth": bandwidth_param}

        super().__init__(
            obs=obs,
            params=params,
            dist_params={},
            dist_kwargs=dist_kwargs,
            distribution=distribution,
            extended=extended,
            norm=norm,
            name=name,
        )
        self.hs3.original_init.update(original_init)


def bw_is_arraylike(bw, allow1d):
    return (
        hasattr(bw, "shape")
        and bw.shape
        and len(bw.shape) > 0
        and (bw.shape[0] is None or (bw.shape[0] > 1 or allow1d))
    )


class KDE1DimGridRepr(BasePDFRepr):
    _implementation = KDE1DimGrid
    hs3_type: Literal["KDE1DimGrid"] = pydantic.Field("KDE1DimGrid", alias="type")

    data: Union[np.ndarray, Serializer.types.DataTypeDiscriminated]
    obs: Optional[SpaceRepr] = None
    bandwidth: Optional[Union[str, float]] = None
    num_grid_points: Optional[int] = None
    binning_method: Optional[str] = None
    kernel: None = None
    padding: Optional[Union[bool, str]] = None
    weights: Optional[Union[np.ndarray, tf.Tensor]] = None
    name: Optional[str] = "GridKDE1DimV1"

    @pydantic.validator("kernel", pre=True)
    def validate_kernel(cls, v):
        if v is not None:
            if v != tfd.Normal:
                raise ValueError(
                    "Kernel must be None for GridKDE1DimV1 to be serialized."
                )
            else:
                v = None
        return v

    @pydantic.root_validator(pre=True)
    def validate_all(cls, values):
        values = dict(values)
        if cls.orm_mode(values):
            for k, v in values["hs3"].original_init.items():
                values[k] = v
        return values


class KDE1DimFFT(KDEHelper, BasePDF, SerializableMixin):
    _N_OBS = 1

    def __init__(
        self,
        data: ztyping.XTypeInput,
        *,
        obs: ztyping.ObsTypeInput | None = None,
        bandwidth: ztyping.ParamTypeInput | str | Callable | None = None,
        kernel: tfd.Distribution = None,
        num_grid_points: int | None = None,
        binning_method: str | None = None,
        support=None,
        fft_method: str | None = None,
        padding: callable | str | bool | None = None,
        weights: np.ndarray | tf.Tensor | None = None,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name: str = "KDE1DimFFT",
    ):
        r"""Kernel Density Estimation is a non-parametric method to approximate the density of given points.

        For a more in-depth explanation, see also in the section about Kernel Density Estimation
        :ref:`sec-fft-kdes`

        .. math::
            f_h(x) =  \frac{1}{nh} \sum_{i=1}^n K\Big(\frac{x-x_i}{h}\Big)

        It is computed by using a convolution of the data with the kernels evaluated at fixed grid points and then
        interpolating between this points to get an estimate for x.

        Args:
            data: |@doc:pdf.kde.init.data| Data sample to approximate
             the density from. The points represent positions of the *kernel*,
             the :math:`x_i`. This is preferrably a ``ZfitData``, but can also
             be an array-like object.

             If the data has weights, they will be taken into account.
             This will change the count of the events, whereas
             weight :math:`w_i` of :math:`x_i` will scale the value of
             :math:`K_i( x_i)`, resulting in a factor of :math:`\frac{w_i}{\sum w_i} `.

             If no weights are given, each kernel will be scaled by the same
             constant :math:`\frac{1}{n_{data}}`. |@docend:pdf.kde.init.data|
            bandwidth: |@doc:pdf.kde.init.bandwidth| Bandwidth of the kernel,
             often also denoted as :math:`h`. For a Gaussian kernel, this
             corresponds to *sigma*. This can be calculated using
             pre-defined options or by specifying a numerical value that is
             broadcastable to *data* -- a scalar or an array-like
             object with the same size as *data*.

             A scalar value is usually referred to as a global bandwidth while
             an array holds local bandwidths |@docend:pdf.kde.init.bandwidth|
            kernel: |@doc:pdf.kde.init.kernel| The kernel is the heart
             of the Kernel Density Estimation, which consists of the sum of
             kernels around each sample point. Therefore, a kernel should represent
             the distribution probability of a single data point as close as
             possible.

             The most widespread kernel is a Gaussian, or Normal, distribution. Due
             to the law of large numbers, the sum of many (arbitrary) random variables
             -- this is the case for most real world observable as they are the result of
             multiple consecutive random effects -- results in a Gaussian distribution.
             However, there are many cases where this assumption is not per-se true. In
             this cases an alternative kernel may offer a better choice.

             Valid choices are callables that return a
             :py:class:`~tensorflow_probability.distribution.Distribution`, such as all distributions
             that belong to the loc-scale family. |@docend:pdf.kde.init.kernel|
            padding: |@doc:pdf.kde.init.padding| KDEs have a peculiar
             weakness: the boundaries, as the outside has a zero density. This makes the KDE
             go down at the bountary as well, as the density approaches zero, no matter what the
             density inside the boundary was.

             There are two ways to circumvent this problem:

               - the best solution: providing a larger dataset than the default space the PDF is used in
               - mirroring the existing data at the boundaries, which is equivalent to a boundary condition
                 with a zero derivative. This is a padding technique and can improve the boundaries.
                 However, one important drawback of this method is to keep in mind that this will actually
                 alter the PDF *to look mirrored*. If the PDF is plotted in a larger range, this becomes
                 clear.

             Possible options are a number (default 0.1) that depicts the fraction of the overall space
             that defines the data mirrored on both sides. For example, for a space from 0 to 5, a value of
             0.3 means that all data in the region of 0 to 1.5 is taken, mirrored around 0 as well as
             all data from 3.5 to 5 and mirrored at 5. The new data will go from -1.5 to 6.5, so the
             KDE is also having a shape outside the desired range. Using it only for the range 0 to 5
             hides this.
             Using a dict, each side separately (or only a single one) can be mirrored, like ``{'lowermirror: 0.1}``
             or ``{'lowermirror: 0.2, 'uppermirror': 0.1}``.
             For more control, a callable that takes data and weights can also be used. |@docend:pdf.kde.init.padding|
            num_grid_points: |@doc:pdf.kde.init.num_grid_points| Number of points in
             the binning grid.

             The data will be binned using the *binning_method* in *num_grid_points*
             and this histogram grid will then be used as kernel points. This has the
             advantage to have a constant computational complexity independent of the data
             size.

             A number from 32 on can already yield good results, while the default is set
             to 1024, creating a fine grid. Lowering the number increases the performance
             at the cost of accuracy. |@docend:pdf.kde.init.num_grid_points|
            binning_method: |@doc:pdf.kde.init.binning_method| Method to be used for
             binning the data. Options are 'linear', 'simple'.

             The data can be binned in the usual way ('simple'), but this is less precise
             for KDEs, where we are interested in the shape of the histogram and smoothing
             it. Therefore, a better suited method, 'linear', is available.

             In normal binnig, each event (or weight) falls into the bin within the bin edges,
             while the neighbouring bins get zero counts from this event.
             In linear binning, the event is split between two bins, proportional to its
             closeness to each bin.

             The 'linear' method provides superior performance, most notably in small (~32)
             grids. |@docend:pdf.kde.init.binning_method|
            support:
            fft_method:
            obs: |@doc:pdf.kde.init.obs| Observable space of the KDE.
             As with any other PDF, this will be used as the default *norm*, but
             does not define the domain of the PDF. Namely, this can be a smaller
             space than *data*, as long as the name of the observable match.
             Using a larger dataset is actually good practice avoiding
             bountary biases, see also :ref:`sec-boundary-bias-and-padding`. |@docend:pdf.kde.init.obs|
            weights: |@doc:pdf.kde.init.weights| Weights of each event
             in *data*, can be None or Tensor-like with shape compatible
             with *data*. Instead of using this parameter, it is preferred
             to use a ``ZfitData`` as *data* that contains weights.
             This will change the count of the events, whereas
             weight :math:`w_i` of :math:`x_i` will scale the value of :math:`K_i( x_i)`,
             resulting in a factor of :math:`\frac{w_i}{\sum w_i} `.

             If no weights are given, each kernel will be scaled by the same
             constant :math:`\frac{1}{n_{data}}`. |@docend:pdf.kde.init.weights|
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: |@doc:model.init.name| Human-readable name
               or label of
               the PDF for better identification.
               Has no programmatical functional purpose as identification. |@docend:model.init.name|
            extended: |@doc:model.init.extended||@docend:model.init.extended|
        """
        original_init = {
            "data": data,
            "bandwidth": bandwidth,
            "num_grid_points": num_grid_points,
            "binning_method": binning_method,
            "support": support,
            "fft_method": fft_method,
            "obs": obs,
            "weights": weights,
            "extended": extended,
            "norm": norm,
            "name": name,
        }
        if isinstance(bandwidth, ZfitParameter):
            raise TypeError("bandwidth cannot be a Parameter for the FFT KDE.")
        if num_grid_points is None:
            num_grid_points = self._default_num_grid_points
        if binning_method is None:
            binning_method = "linear"
        if fft_method is None:
            fft_method = "conv1d"
        if kernel is None:
            kernel = tfd.Normal

        if padding is None:
            padding = self._default_padding
        if obs is None:
            if not isinstance(data, ZfitData) or not data.space.has_limits:
                raise ValueError(
                    "obs can only be None if data is ZfitData with limits."
                )
            else:
                obs = data.space
        data, size, weights, _ = self._convert_init_data_weights_size(
            data, weights, padding=padding, limits=obs.limits, bandwidth=bandwidth
        )
        self._padding = padding

        bandwidth, bandwidth_param = self._convert_input_bandwidth(
            bandwidth=bandwidth,
            data=data,
            padding=False,
            kernel=kernel,
            support=support,
            fft_method=fft_method,
            name=name,
            obs=obs,
            weights=weights,
        )
        num_grid_points = tf.minimum(
            tf.cast(size, ztypes.int), tf.constant(num_grid_points, ztypes.int)
        )
        check_bw_grid_shapes(bandwidth, n_grid=num_grid_points)
        self._num_grid_points = num_grid_points
        self._binning_method = binning_method
        self._fft_method = fft_method
        self._data = data

        self._bandwidth = bandwidth

        params = {"bandwidth": self._bandwidth}
        super().__init__(
            obs=obs, name=name, params=params, extended=extended, norm=norm
        )
        self._kernel = kernel
        self._weights = weights
        if support is None:
            area = znp.reshape(self.space.area(), ())
            if area is not None:
                support = area * 1.2
        self._support = support
        self._grid = None
        self._grid_data = None

        self._grid = binning_util.generate_1d_grid(
            self._data, num_grid_points=self._num_grid_points
        )
        self._grid_data = binning_util.bin_1d(
            self._binning_method, self._data, self._grid, self._weights
        )
        self._grid_estimations = convolution_util.convolve_1d_data_with_kernel(
            self._kernel,
            self._bandwidth,
            self._grid_data,
            self._grid,
            self._support,
            self._fft_method,
        )
        self.hs3.original_init.update(original_init)

    def _unnormalized_pdf(self, x):
        x = z.unstack_x(x)
        x_min = self._grid[0]
        x_max = self._grid[-1]

        value = tfp.math.interp_regular_1d_grid(x, x_min, x_max, self._grid_estimations)
        value.set_shape(x.shape)
        return value


class KDE1DimFFTRepr(BasePDFRepr):
    _implementation = KDE1DimFFT
    hs3_type: Literal["KDE1DimFFT"] = pydantic.Field("KDE1DimFFT", alias="type")

    data: Union[np.ndarray, Serializer.types.DataTypeDiscriminated]
    obs: Optional[SpaceRepr] = None
    bandwidth: Optional[Union[str, float]] = None
    num_grid_points: Optional[int] = None
    binning_method: Optional[str] = None
    kernel: None = None
    support: Optional[float] = None
    fft_method: Optional[str] = None
    padding: Optional[Union[bool, str]] = None
    weights: Optional[Union[np.ndarray, tf.Tensor]] = None
    name: Optional[str] = "KDE1DimFFT"

    @pydantic.validator("kernel", pre=True)
    def validate_kernel(cls, v):
        if v is not None:
            if v != tfd.Normal:
                raise ValueError(
                    "Kernel must be None for GridKDE1DimV1 to be serialized."
                )
            else:
                v = None
        return v

    @pydantic.root_validator(pre=True)
    def validate_all(cls, values):
        values = dict(values)
        if cls.orm_mode(values):
            for k, v in values["hs3"].original_init.items():
                values[k] = v
        return values


class KDE1DimISJ(KDEHelper, BasePDF, SerializableMixin):
    _N_OBS = 1

    def __init__(
        self,
        data: ztyping.XTypeInput,
        *,
        obs: ztyping.ObsTypeInput | None = None,
        padding: callable | str | bool | None = None,
        num_grid_points: int | None = None,
        binning_method: str | None = None,
        weights: np.ndarray | tf.Tensor | None = None,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name: str = "KDE1DimISJ",
    ):
        r"""Kernel Density Estimation is a non-parametric method to approximate the density of given points.

        For a more in-depth explanation, see also in the section about Kernel Density Estimation
        :ref:`sec-isj-kde`

        .. math::

            f_h(x) =  \frac{1}{nh} \sum_{i=1}^n K\Big(\frac{x-x_i}{h}\Big)

        The bandwidth is computed by using a trick described in a paper by Botev et al. that uses the fact,
        that the Kernel Density Estimation
        with a Gaussian Kernel is a solution to the Heat Equation.

        Args:
            data: |@doc:pdf.kde.init.data| Data sample to approximate
             the density from. The points represent positions of the *kernel*,
             the :math:`x_i`. This is preferrably a ``ZfitData``, but can also
             be an array-like object.

             If the data has weights, they will be taken into account.
             This will change the count of the events, whereas
             weight :math:`w_i` of :math:`x_i` will scale the value of
             :math:`K_i( x_i)`, resulting in a factor of :math:`\frac{w_i}{\sum w_i} `.

             If no weights are given, each kernel will be scaled by the same
             constant :math:`\frac{1}{n_{data}}`. |@docend:pdf.kde.init.data|
            padding: |@doc:pdf.kde.init.padding| KDEs have a peculiar
             weakness: the boundaries, as the outside has a zero density. This makes the KDE
             go down at the bountary as well, as the density approaches zero, no matter what the
             density inside the boundary was.

             There are two ways to circumvent this problem:

               - the best solution: providing a larger dataset than the default space the PDF is used in
               - mirroring the existing data at the boundaries, which is equivalent to a boundary condition
                 with a zero derivative. This is a padding technique and can improve the boundaries.
                 However, one important drawback of this method is to keep in mind that this will actually
                 alter the PDF *to look mirrored*. If the PDF is plotted in a larger range, this becomes
                 clear.

             Possible options are a number (default 0.1) that depicts the fraction of the overall space
             that defines the data mirrored on both sides. For example, for a space from 0 to 5, a value of
             0.3 means that all data in the region of 0 to 1.5 is taken, mirrored around 0 as well as
             all data from 3.5 to 5 and mirrored at 5. The new data will go from -1.5 to 6.5, so the
             KDE is also having a shape outside the desired range. Using it only for the range 0 to 5
             hides this.
             Using a dict, each side separately (or only a single one) can be mirrored, like ``{'lowermirror: 0.1}``
             or ``{'lowermirror: 0.2, 'uppermirror': 0.1}``.
             For more control, a callable that takes data and weights can also be used. |@docend:pdf.kde.init.padding|
            num_grid_points: |@doc:pdf.kde.init.num_grid_points| Number of points in
             the binning grid.

             The data will be binned using the *binning_method* in *num_grid_points*
             and this histogram grid will then be used as kernel points. This has the
             advantage to have a constant computational complexity independent of the data
             size.

             A number from 32 on can already yield good results, while the default is set
             to 1024, creating a fine grid. Lowering the number increases the performance
             at the cost of accuracy. |@docend:pdf.kde.init.num_grid_points|
            binning_method: |@doc:pdf.kde.init.binning_method| Method to be used for
             binning the data. Options are 'linear', 'simple'.

             The data can be binned in the usual way ('simple'), but this is less precise
             for KDEs, where we are interested in the shape of the histogram and smoothing
             it. Therefore, a better suited method, 'linear', is available.

             In normal binnig, each event (or weight) falls into the bin within the bin edges,
             while the neighbouring bins get zero counts from this event.
             In linear binning, the event is split between two bins, proportional to its
             closeness to each bin.

             The 'linear' method provides superior performance, most notably in small (~32)
             grids. |@docend:pdf.kde.init.binning_method|
            obs: |@doc:pdf.kde.init.obs| Observable space of the KDE.
             As with any other PDF, this will be used as the default *norm*, but
             does not define the domain of the PDF. Namely, this can be a smaller
             space than *data*, as long as the name of the observable match.
             Using a larger dataset is actually good practice avoiding
             bountary biases, see also :ref:`sec-boundary-bias-and-padding`. |@docend:pdf.kde.init.obs|
            weights: |@doc:pdf.kde.init.weights| Weights of each event
             in *data*, can be None or Tensor-like with shape compatible
             with *data*. Instead of using this parameter, it is preferred
             to use a ``ZfitData`` as *data* that contains weights.
             This will change the count of the events, whereas
             weight :math:`w_i` of :math:`x_i` will scale the value of :math:`K_i( x_i)`,
             resulting in a factor of :math:`\frac{w_i}{\sum w_i} `.

             If no weights are given, each kernel will be scaled by the same
             constant :math:`\frac{1}{n_{data}}`. |@docend:pdf.kde.init.weights|
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: |@doc:model.init.name| Human-readable name
               or label of
               the PDF for better identification.
               Has no programmatical functional purpose as identification. |@docend:model.init.name|
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
        """
        original_init = {
            "data": data,
            "weights": weights,
            "num_grid_points": num_grid_points,
            "binning_method": binning_method,
            "obs": obs,
            "padding": padding,
            "name": name,
            "norm": norm,
            "extended": extended,
        }
        if num_grid_points is None:
            num_grid_points = self._default_num_grid_points
        if binning_method is None:
            binning_method = "linear"
        if padding is None:
            padding = self._default_padding
        if obs is None:
            if not isinstance(data, ZfitData) or not data.space.has_limits:
                raise ValueError(
                    "obs can only be None if data is ZfitData with limits."
                )
            else:
                obs = data.space
        data, size, weights, _ = self._convert_init_data_weights_size(
            data, weights, padding=padding, limits=obs.limits, bandwidth=None
        )
        self._padding = padding

        num_grid_points = tf.minimum(
            tf.cast(size, ztypes.int), tf.constant(num_grid_points, ztypes.int)
        )
        self._num_grid_points = num_grid_points
        self._binning_method = binning_method
        self._data = tf.convert_to_tensor(data, ztypes.float)
        self._weights = weights
        self._grid = None
        self._grid_data = None

        (
            self._bandwidth,
            self._grid_estimations,
            self._grid,
        ) = isj_util.calculate_bandwidth_and_density(
            self._data, self._num_grid_points, self._binning_method, self._weights
        )

        params = {}
        super().__init__(
            obs=obs, name=name, params=params, extended=extended, norm=norm
        )
        self.hs3.original_init.update(original_init)

    def _unnormalized_pdf(self, x):
        x = z.unstack_x(x)
        x_min = self._grid[0]
        x_max = self._grid[-1]

        value = tfp.math.interp_regular_1d_grid(x, x_min, x_max, self._grid_estimations)
        value.set_shape(x.shape)
        return value


class KDE1DimISJRepr(BasePDFRepr):
    _implementation = KDE1DimISJ
    hs3_type: Literal["KDE1DimISJ"] = pydantic.Field("KDE1DimISJ", alias="type")

    data: Union[np.ndarray, Serializer.types.DataTypeDiscriminated]
    obs: Optional[SpaceRepr] = None
    bandwidth: Optional[Union[str, float]] = None
    num_grid_points: Optional[int] = None
    binning_method: Optional[str] = None
    kernel: None = None
    padding: Optional[Union[bool, str]] = None
    weights: Optional[Union[np.ndarray, tf.Tensor]] = None
    name: Optional[str] = "KDE1DimISJ"

    @pydantic.validator("kernel", pre=True)
    def validate_kernel(cls, v):
        if v is not None:
            if v != tfd.Normal:
                raise ValueError(
                    "Kernel must be None for GridKDE1DimV1 to be serialized."
                )
            else:
                v = None
        return v

    @pydantic.root_validator(pre=True)
    def validate_all(cls, values):
        values = dict(values)
        if cls.orm_mode(values):
            for k, v in values["hs3"].original_init.items():
                values[k] = v
        return values



#  Copyright (c) 2022 zfit
import pickle

import numpy as np
import pytest

import zfit._data.unbinneddata
import zfit._variables.axis as axis
import zfit.z.numpy as znp


@pytest.mark.skip
def test_basic():
    n = 1000
    nobs = 3
    axes = axis.SpaceV2(
        [axis.UnbinnedAxis("x"), axis.UnbinnedAxis("y"), axis.UnbinnedAxis("z")]
    )
    tensor = znp.random.uniform(size=(n, nobs))
    weights = znp.random.uniform(size=(n,))
    data = zfit._data.unbinneddata.UnbinnedData(data=tensor, axes=axes, weights=weights)

    pickled = pickle.dumps(data)
    unpickled = pickle.loads(pickled)
    # assert data.axes == unpickled.axes
    np.testing.assert_array_equal(data.weights, unpickled.weights)
    np.testing.assert_array_equal(data.data, unpickled.data)

    array = data["y"]
    np.testing.assert_array_equal(array, tensor[:, 1])

    # af = asdf.AsdfFile(data.obj_to_repr())
    # with io.BytesIO(b'file') as f:
    #     af.write_to(f)



#  Copyright (c) 2023 zfit

import numpy as np
import pytest


@pytest.mark.parametrize(
    "bandwidth", [None, 0.1, np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0])]
)
@pytest.mark.parametrize(
    "weights", [None, np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0])]
)
def test_padreflect_data_weights_1dim(weights, bandwidth):
    from zfit import Space
    from zfit.models.kde import padreflect_data_weights_1dim

    testarray = np.array([0.1, 0.5, -2, 5, 6.7, -1.2, 1.2, 7])
    orig_bw = bandwidth

    data, w, bw = padreflect_data_weights_1dim(
        testarray, mode=0.1, weights=weights, bandwidth=bandwidth
    )
    true_data = np.array([-2.8, -2, -2, -1.2, 0.1, 0.5, 1.2, 5, 6.7, 7, 7, 7.3])

    np.testing.assert_allclose(true_data, np.sort(data))
    true_w = np.sort(np.array([6, 3, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 8, 5]))
    if weights is not None:
        np.testing.assert_allclose(true_w, np.sort(w))
    if bandwidth is not None:
        if isinstance(orig_bw, np.ndarray):
            true_bw = true_w.copy()
            np.testing.assert_allclose(true_bw, np.sort(bw))
        else:
            assert bw == orig_bw

    data, w, bw = padreflect_data_weights_1dim(
        testarray,
        mode={"lowermirror": 0.1, "uppermirror": 0.1},
        weights=weights,
        bandwidth=bandwidth,
    )
    true_data = np.array([-2.8, -2, -2, -1.2, 0.1, 0.5, 1.2, 5, 6.7, 7, 7, 7.3])
    np.testing.assert_allclose(true_data, np.sort(data))
    true_w = np.sort(np.array([6, 3, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 8, 5]))
    if weights is not None:
        np.testing.assert_allclose(true_w, np.sort(w))
    if bandwidth is not None:
        if isinstance(orig_bw, np.ndarray):
            true_bw = true_w.copy()
            np.testing.assert_allclose(true_bw, np.sort(bw))
        else:
            assert bw == orig_bw

    data, w, bw = padreflect_data_weights_1dim(
        testarray,
        mode={"lowermirror": 0.1, "uppermirror": 0.3},
        weights=weights,
        bandwidth=bandwidth,
    )
    true_data = np.sort(
        np.array([-2.8, -2, -2, -1.2, 0.1, 0.5, 1.2, 5, 6.7, 7, 7, 7.3, 9])
    )
    np.testing.assert_allclose(true_data, np.sort(data))
    true_w = np.sort(np.array([6, 3, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 8, 5, 4]))
    if weights is not None:
        np.testing.assert_allclose(true_w, np.sort(w))

    if bandwidth is not None:
        if isinstance(orig_bw, np.ndarray):
            true_bw = true_w.copy()
            np.testing.assert_allclose(true_bw, np.sort(bw))
        else:
            assert bw == orig_bw

    data, w, bw = padreflect_data_weights_1dim(
        testarray,
        mode={"lowermirror": 0.1},
        weights=weights,
        bandwidth=bandwidth,
    )
    true_data = np.sort(np.array([-2.8, -2, -2, -1.2, 0.1, 0.5, 1.2, 5, 6.7, 7]))
    np.testing.assert_allclose(true_data, np.sort(data))
    true_w = np.sort(
        np.array(
            [
                6,
                3,
                1.0,
                2.0,
                3.0,
                4.0,
                5.0,
                6.0,
                7.0,
                8.0,
            ]
        )
    )
    if weights is not None:
        np.testing.assert_allclose(true_w, np.sort(w))

    if bandwidth is not None:
        if isinstance(orig_bw, np.ndarray):
            true_bw = true_w.copy()
            np.testing.assert_allclose(true_bw, np.sort(bw))
        else:
            assert bw == orig_bw

    data, w, bw = padreflect_data_weights_1dim(
        testarray,
        mode={"lowermirror": 0.3},
        weights=weights,
        bandwidth=bandwidth,
    )
    true_data = np.sort(
        np.array([-4.1, -4.5, -2.8, -2, -2, -1.2, 0.1, 0.5, 1.2, 5, 6.7, 7])
    )
    np.testing.assert_allclose(true_data, np.sort(data))
    true_w = np.sort(
        np.array(
            [
                1,
                2,
                6,
                3,
                1.0,
                2.0,
                3.0,
                4.0,
                5.0,
                6.0,
                7.0,
                8.0,
            ]
        )
    )
    if weights is not None:
        np.testing.assert_allclose(true_w, np.sort(w))
    if bandwidth is not None:
        if isinstance(orig_bw, np.ndarray):
            true_bw = true_w.copy()
            np.testing.assert_allclose(true_bw, np.sort(bw))
        else:
            assert bw == orig_bw

    data, w, bw = padreflect_data_weights_1dim(
        testarray,
        mode={"uppermirror": 0.3},
        weights=weights,
        bandwidth=bandwidth,
    )
    true_data = np.sort(np.array([-2, -1.2, 0.1, 0.5, 1.2, 5, 6.7, 7, 7, 7.3, 9]))
    np.testing.assert_allclose(true_data, np.sort(data))
    true_w = np.sort(np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 8, 5, 4]))
    if weights is not None:
        np.testing.assert_allclose(true_w, np.sort(w))
    if bandwidth is not None:
        if isinstance(orig_bw, np.ndarray):
            true_bw = true_w.copy()
            np.testing.assert_allclose(true_bw, np.sort(bw))
        else:
            assert bw == orig_bw

    testobs = Space("testSpace", limits=(-2, 7.5))
    data, w, bw = padreflect_data_weights_1dim(
        testarray, mode=0.1, limits=testobs.limits, weights=weights, bandwidth=bandwidth
    )
    true_data = np.sort(
        np.array([-2.8, -2, 0.1, 0.5, -2, 5, 6.7, -1.2, 1.2, 7, 8.0, 8.3])
    )
    np.testing.assert_allclose(true_data, np.sort(data))
    true_w = np.sort(np.array([3, 6, 1, 2, 3, 4, 5, 6, 7, 8, 5, 8]))
    if weights is not None:
        np.testing.assert_allclose(true_w, np.sort(w))
    if bandwidth is not None:
        if isinstance(orig_bw, np.ndarray):
            true_bw = true_w.copy()
            np.testing.assert_allclose(true_bw, np.sort(bw))
        else:
            assert bw == orig_bw

    data, w, bw = padreflect_data_weights_1dim(
        testarray, mode=0.3, limits=testobs.limits, weights=weights, bandwidth=bandwidth
    )
    true_data = np.sort(
        np.array(
            [-4.5, -4.1, -2.8, -2, 0.1, 0.5, -2, 5, 6.7, -1.2, 1.2, 7, 8.0, 8.3, 10]
        )
    )
    np.testing.assert_allclose(true_data, np.sort(data))
    true_w = np.sort(np.array([1, 2, 3, 6, 1, 2, 3, 4, 5, 6, 7, 8, 4, 5, 8]))
    if weights is not None:
        np.testing.assert_allclose(true_w, np.sort(w))

    if bandwidth is not None:
        if isinstance(orig_bw, np.ndarray):
            true_bw = true_w.copy()
            np.testing.assert_allclose(true_bw, np.sort(bw))
        else:
            assert bw == orig_bw

    data, w, bw = padreflect_data_weights_1dim(
        testarray,
        mode={"lowermirror": 0.3},
        limits=testobs.limits,
        weights=weights,
        bandwidth=bandwidth,
    )
    true_data = np.sort(
        np.array([-4.5, -4.1, -2.8, -2, 0.1, 0.5, -2, 5, 6.7, -1.2, 1.2, 7])
    )
    np.testing.assert_allclose(true_data, np.sort(data))
    true_w = np.sort(np.array([1, 2, 3, 6, 1, 2, 3, 4, 5, 6, 7, 8]))
    if weights is not None:
        np.testing.assert_allclose(true_w, np.sort(w))
    if bandwidth is not None:
        if isinstance(orig_bw, np.ndarray):
            true_bw = true_w.copy()
            np.testing.assert_allclose(true_bw, np.sort(bw))
        else:
            assert bw == orig_bw

    data, w, bw = padreflect_data_weights_1dim(
        testarray,
        mode={"uppermirror": 0.3},
        limits=testobs.limits,
        weights=weights,
        bandwidth=bandwidth,
    )
    true_data = np.sort(np.array([0.1, 0.5, -2, 5, 6.7, -1.2, 1.2, 7, 8.0, 8.3, 10]))
    np.testing.assert_allclose(true_data, np.sort(data))
    true_w = np.sort(np.array([1, 2, 3, 4, 5, 6, 7, 8, 4, 5, 8]))
    if weights is not None:
        np.testing.assert_allclose(true_w, np.sort(w))

    with pytest.raises(ValueError):
        data, w, bw = padreflect_data_weights_1dim(
            testarray,
            mode={"uper": 0.3},
            weights=weights,
            bandwidth=bandwidth,
        )
    with pytest.raises(ValueError):
        data, w, bw = padreflect_data_weights_1dim(
            testarray,
            mode={"uppermirror": 0.3, "lowe": 0.321},
            weights=weights,
            bandwidth=bandwidth,
        )



#  Copyright (c) 2023 zfit

from __future__ import annotations

from typing import TYPE_CHECKING

from ..util.ztyping import ExtendedInputType, NormInputType

if TYPE_CHECKING:
    pass

from collections.abc import Mapping

import functools
import warnings

import tensorflow as tf

import zfit.z.numpy as znp

from .. import z
from ..core.interfaces import (
    ZfitIndependentParameter,
    ZfitPDF,
    ZfitSpace,
    ZfitParameter,
)
from ..core.basepdf import BasePDF
from ..core.parameter import set_values
from ..core.space import combine_spaces, convert_to_space, supports
from ..util.exception import WorkInProgressError
from ..util.warnings import warn_experimental_feature
from .functor import BaseFunctor


class ConditionalPDFV1(BaseFunctor):
    @warn_experimental_feature
    def __init__(
        self,
        pdf: ZfitPDF,
        cond: Mapping[ZfitIndependentParameter, ZfitSpace],
        *,
        name: str = "ConditionalPDF",
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        use_vectorized_map: bool = False,
        sample_with_replacement: bool = True,
    ) -> None:
        """EXPERIMENTAL! Implementation of a Conditional PDF, rather slow and for research purpose.

        As an example, a Gaussian is wrapped in order to make 'sigma' conditional.

        .. jupyter-execute::



        Args:
            pdf: PDF that will be wrapped. Convert one or several parameters of *pdf* to a conditional
                parameter, meaning that the parameter *param* in the ``cond`` mapping will now be
                determined by the data in the ``Space``, the value of the ``cond``.
            cond: Mapping of parameter to input data.
            name: |@doc:model.init.name| Human-readable name
               or label of
               the PDF for better identification.
               Has no programmatical functional purpose as identification. |@docend:model.init.name|
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:model.init.norm| The normalization of the PDF. |@docend:model.init.norm|
            use_vectorized_map ():
            sample_with_replacement ():
        """
        # TODO: add to serializer, see below repr for problem
        # original_init = {'pdf': pdf, 'cond': cond, 'name': name, 'extended': extended, 'norm': norm,
        #                  'use_vectorized_map': use_vectorized_map, 'sample_with_replacement': sample_with_replacement}
        self._sample_with_replacement = sample_with_replacement
        self._use_vectorized_map = use_vectorized_map
        self._cond, cond_obs = self._check_input_cond(cond)
        obs = pdf.space * cond_obs
        super().__init__(pdfs=pdf, obs=obs, name=name, extended=extended, norm=norm)
        # self.hs3.original_init.update(original_init)  # TODO: add to serializer

    @property
    def cond(self) -> dict[ZfitIndependentParameter, ZfitSpace]:
        return self._cond

    def _check_input_cond(self, cond):
        spaces = []
        for param, obs in cond.items():
            if not isinstance(param, ZfitIndependentParameter):
                raise TypeError(f"parameter {param} not a ZfitIndependentParameter")
            spaces.append(convert_to_space(obs))
        return cond, combine_spaces(*spaces)

    @supports(norm=True, multiple_limits=True)
    @z.function(wraps="conditional_pdf")
    def _pdf(self, x, norm):
        pdf = self.pdfs[0]
        param_x_indices = {
            p: x.obs.index(p_space.obs[0]) for p, p_space in self._cond.items()
        }
        x_values = x.value()

        if self._use_vectorized_map:
            tf_map = tf.vectorized_map
        else:
            output_signature = tf.TensorSpec(
                shape=(1, *x_values.shape[1:-1]), dtype=self.dtype
            )
            tf_map = functools.partial(tf.map_fn, fn_output_signature=output_signature)

        # TODO: reset parameters?

        def eval_pdf(cond_and_data):
            x_pdf = cond_and_data[None, ..., : pdf.n_obs]
            for param, index in param_x_indices.items():
                param.assign(cond_and_data[..., index])
            return pdf.pdf(x_pdf, norm=norm)

        params = tuple(param_x_indices.keys())
        with set_values(params, params):
            probs = tf_map(eval_pdf, x_values)
        probs = probs[:, 0]  # removing stack dimension, implicitly in map_fn
        return probs

    def _get_params(
        self,
        floating: bool | None = True,
        is_yield: bool | None = None,
        extract_independent: bool | None = True,
    ) -> set[ZfitParameter]:
        params = super()._get_params(floating, is_yield, extract_independent)
        params -= set(self._cond)
        return params

    @z.function(wraps="conditional_pdf")
    def _single_hook_integrate(self, limits, norm, x, options):
        from zfit import run

        if not run.get_graph_mode():
            warnings.warn(
                "Using the Conditional PDF in eager mode (no jit) maybe gets stuck.",
                RuntimeWarning,
            )

        param_x_indices = {
            p: x.obs.index(p_space.obs[0]) for p, p_space in self._cond.items()
        }
        x_values = x.value()
        pdf = self.pdfs[0]

        if self._use_vectorized_map:
            tf_map = tf.vectorized_map
        else:
            output_signature = tf.TensorSpec(
                shape=(1, *x_values.shape[1:-1]), dtype=self.dtype
            )
            tf_map = functools.partial(tf.map_fn, fn_output_signature=output_signature)

        @z.function(wraps="vectorized_map")
        def eval_int(values):
            for param, index in param_x_indices.items():
                param.assign(values[..., index])

            return pdf.integrate(limits=limits, norm=norm, options=options)

        integrals = tf_map(eval_int, x_values)
        integrals = integrals[:, 0]  # removing stack dimension, implicitly in map_fn
        return integrals

    @z.function(wraps="conditional_pdf")
    def _single_hook_sample(self, n, limits, x):
        tf.assert_equal(
            n,
            x.nevents,
            message="Different number of n requested than x given for "
            "conditional sampling. Needs to agree",
        )

        param_x_indices = {
            p: x.obs.index(p_space.obs[0]) for p, p_space in self._cond.items()
        }
        x_values = x.value()
        # if self._sample_with_replacement:
        #     x_values = z.random.sample_with_replacement(x_values, axis=0, sample_shape=(n,))
        pdf = self.pdfs[0]

        if self._use_vectorized_map:
            tf_map = tf.vectorized_map
        else:
            output_signature = tf.TensorSpec(shape=(1, pdf.n_obs), dtype=self.dtype)
            tf_map = functools.partial(tf.map_fn, fn_output_signature=output_signature)

        def eval_sample(values):
            for param, index in param_x_indices.items():
                param.assign(values[..., index])

            return pdf.sample(n=1, limits=limits).value()

        sample_rnd = tf_map(eval_sample, x_values)[..., 0]
        sample = znp.concatenate([sample_rnd, x_values], axis=-1)
        return sample

    def copy(self, **override_parameters) -> BasePDF:
        raise WorkInProgressError(
            "Currently copying not possible. " "Use `set_yield` to set a yield inplace."
        )


# NOT working, logic wrong: the parameter of Gauss is not added to overall variables...
# class ConditionalPDFV1Repr(BasePDFRepr):
#     _implementation = ConditionalPDFV1
#     hs3_type: Literal["ConditionalPDFV1"] = pydantic.Field("ConditionalPDFV1", alias="type")
#
#     pdf: List[Serializer.types.PDFTypeDiscriminated]
#     cond: Dict[Serializer.types.ParamTypeDiscriminated, Union[SpaceRepr, Tuple[str]]]
#     obs: Optional[Union[SpaceRepr, Tuple[str]]] = None
#     extended: Serializer.types.ParamInputTypeDiscriminated = None
#
#     #
#     @pydantic.root_validator(pre=True)
#     def validate_all(cls, values):
#         if cls.orm_mode(values):
#             values = dict(values)
#             for k, v in values['hs3'].original_init.items():
#                 values[k] = v
#             values['pdf'] = [values['pdf']]
#             values['obs'] = values['space']
#         return values
#
#     def _to_orm(self, init):
#         init = dict(init)
#         init['pdf'] = init['pdf'][0]
#         out = super()._to_orm(init)
#         return out



#  Copyright (c) 2022 zfit

import tensorflow as tf

from .core.constraint import (
    GaussianConstraint,
    PoissonConstraint,
    SimpleConstraint,
    LogNormalConstraint,
)
from .util import ztyping

__all__ = [
    "nll_gaussian",
    "SimpleConstraint",
    "GaussianConstraint",
    "PoissonConstraint",
    "LogNormalConstraint",
]


def nll_gaussian(
    params: ztyping.ParamTypeInput,
    observation: ztyping.NumericalScalarType,
    uncertainty: ztyping.NumericalScalarType,
) -> tf.Tensor:
    """Return negative log likelihood graph for gaussian constraints on a list of parameters.

    Args:
        params: The parameters to constraint.
        observation: observed values of the parameter.
        uncertainty: Uncertainties or covariance/error.
            matrix of the observed values. Can either be a single value, a list of values, an array or a tensor.
    Returns:
        The constraint object.
    Raises:
        ShapeIncompatibleError: if params, mu and sigma don't have the same size.
    """

    return GaussianConstraint(
        params=params, observation=observation, uncertainty=uncertainty
    )



#  Copyright (c) 2023 zfit

import mplhep
import numpy as np

import zfit

mplhep.style.use("LHCb2")
import matplotlib.pyplot as plt

# create space
obs = zfit.Space("x", limits=(-10, 10))

# parameters
mu = zfit.Parameter("mu", 1.0, -4, 6)
sigma = zfit.Parameter("sigma", 1.0, 0.1, 10)
lambd = zfit.Parameter("lambda", -0.06, -1, -0.01)
frac = zfit.Parameter("fraction", 0.3, 0, 1)

# model building, pdf creation
gauss = zfit.pdf.Gauss(mu=mu, sigma=sigma, obs=obs)
exponential = zfit.pdf.Exponential(lambd, obs=obs)
model = zfit.pdf.SumPDF([gauss, exponential], fracs=frac)

# data
n_sample = 5000

exp_data = exponential.sample(n=n_sample * (1 - frac)).numpy()

gauss_data = gauss.sample(n=n_sample * frac).numpy()

data = model.create_sampler(n_sample, limits=obs)
data.resample()

mu.set_value(0.5)
sigma.set_value(1.2)
lambd.set_value(-0.05)
frac.set_value(0.07)

# plot the data
data_np = data["x"].numpy()
n_bins = 50

plot_scaling = n_sample / n_bins * obs.area()

x = np.linspace(-10, 10, 1000)
lower, upper = obs.limit1d
edges = np.linspace(lower, upper, n_bins + 1)


def plot_pdf(title):
    plt.figure()
    plt.title(title)
    y = model.pdf(x).numpy()
    y_gauss = (gauss.pdf(x) * frac).numpy()
    y_exp = (exponential.pdf(x) * (1 - frac)).numpy()
    plt.plot(x, y * plot_scaling, label="Sum - Model")
    plt.plot(x, y_gauss * plot_scaling, label="Gauss - Signal")
    plt.plot(x, y_exp * plot_scaling, label="Exp - Background")
    mplhep.histplot(
        np.histogram(data_np, bins=edges),
        yerr=True,
        color="black",
        histtype="errorbar",
    )
    plt.ylabel("Counts")
    plt.xlabel("$B_{mass}$")


# plot the pdf BEFORE fitting
plot_pdf("Before fitting")
# create NLL
nll = zfit.loss.UnbinnedNLL(model=model, data=data)

# create a minimizer
minimizer = zfit.minimize.Minuit()
result = minimizer.minimize(nll)

# do the error calculations, here with minos
param_errors, _ = result.errors()

plot_pdf(title="After fitting")
plt.legend()

# uncomment to display plots
# plt.show()



#  Copyright (c) 2023 zfit
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import zfit

import tensorflow as tf

import zfit
import zfit.z.numpy as znp
from zfit import z
from .binned_functor import BaseBinnedFunctorPDF
from ..core.interfaces import ZfitPDF, ZfitSpace
from ..util import ztyping
from ..util.warnings import warn_advanced_feature


class BinnedFromUnbinnedPDF(BaseBinnedFunctorPDF):
    def __init__(
        self,
        pdf: ZfitPDF,
        space: ZfitSpace,
        extended: ztyping.ExtendedInputType = None,
        norm: ztyping.NormInputType = None,
    ) -> None:
        """Create a binned pdf from an unbinned pdf binning in *space*.

        Args:
            pdf: The unbinned pdf to be binned.
            space: |@doc:pdf.init.obs| Observables of the
               model. This will be used as the default space of the PDF and,
               if not given explicitly, as the normalization range.

               The default space is used for example in the sample method: if no
               sampling limits are given, the default space is used.

               The observables are not equal to the domain as it does not restrict or
               truncate the model outside this range. |@docend:pdf.init.obs|
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
        """
        if pdf.is_extended:
            if extended is not None:
                warn_advanced_feature(
                    f"PDF {pdf} is already extended, but extended also given {extended}. Will"
                    f" use the given yield.",
                    identifier="extend_wrapped_extended",
                )
            else:
                extended = pdf.get_yield()
        if not isinstance(space, ZfitSpace):
            try:
                space = pdf.space.with_binning(space)
            except Exception as error:
                raise ValueError(
                    f"Could not create space {space} from pdf {pdf} with binning {space}"
                ) from error
        super().__init__(
            obs=space,
            extended=extended,
            norm=norm,
            models=pdf,
            params={},
            name="BinnedFromUnbinnedPDF",
        )
        self.pdfs = self.models

    # def _get_params(self, floating: bool | None = True, is_yield: bool | None = None,
    #                 extract_independent: bool | None = True) -> set[ZfitParameter]:
    #     params = super()._get_params(floating=floating, is_yield=is_yield, extract_independent=extract_independent)
    #     daughter_params = self.pdfs[0].get_params(floating=floating, is_yield=is_yield,
    #                                               extract_independent=extract_independent)
    #     return daughter_params | params

    @z.function
    def _rel_counts(self, x, norm):
        pdf = self.pdfs[0]
        edges = [znp.array(edge) for edge in self.axes.edges]
        edges_flat = [znp.reshape(edge, [-1]) for edge in edges]
        lowers = [edge[:-1] for edge in edges_flat]
        uppers = [edge[1:] for edge in edges_flat]
        lowers_meshed = znp.meshgrid(*lowers, indexing="ij")
        uppers_meshed = znp.meshgrid(*uppers, indexing="ij")
        shape = tf.shape(lowers_meshed[0])
        lowers_meshed_flat = [
            znp.reshape(lower_mesh, [-1]) for lower_mesh in lowers_meshed
        ]
        uppers_meshed_flat = [
            znp.reshape(upper_mesh, [-1]) for upper_mesh in uppers_meshed
        ]
        lower_flat = znp.stack(lowers_meshed_flat, axis=-1)
        upper_flat = znp.stack(uppers_meshed_flat, axis=-1)
        options = {"type": "bins"}

        @z.function
        def integrate_one(limits):
            l, u = tf.unstack(limits)
            limits_space = zfit.Space(obs=self.obs, limits=[l, u])
            return pdf.integrate(limits_space, norm=False, options=options)

        limits = znp.stack([lower_flat, upper_flat], axis=1)
        from zfit import run

        try:
            if run.executing_eagerly():
                raise TypeError("Just stearing the eager execution")
            values = tf.vectorized_map(integrate_one, limits)[:, 0]
        except (ValueError, TypeError):
            with run.aquire_cpu(-1) as cpus:
                values = tf.map_fn(integrate_one, limits, parallel_iterations=len(cpus))
        values = znp.reshape(values, shape)
        if norm:
            values /= pdf.normalization(norm)
        return values

    @z.function(wraps="model_binned")
    def _counts(self, x, norm):
        pdf = self.pdfs[0]
        edges = [znp.array(edge) for edge in self.axes.edges]
        edges_flat = [znp.reshape(edge, [-1]) for edge in edges]
        lowers = [edge[:-1] for edge in edges_flat]
        uppers = [edge[1:] for edge in edges_flat]
        lowers_meshed = znp.meshgrid(*lowers, indexing="ij")
        uppers_meshed = znp.meshgrid(*uppers, indexing="ij")
        shape = tf.shape(lowers_meshed[0])
        lowers_meshed_flat = [
            znp.reshape(lower_mesh, [-1]) for lower_mesh in lowers_meshed
        ]
        uppers_meshed_flat = [
            znp.reshape(upper_mesh, [-1]) for upper_mesh in uppers_meshed
        ]
        lower_flat = znp.stack(lowers_meshed_flat, axis=-1)
        upper_flat = znp.stack(uppers_meshed_flat, axis=-1)
        options = {"type": "bins"}

        if pdf.is_extended:

            @z.function
            def integrate_one(limits):
                l, u = tf.unstack(limits)
                limits_space = zfit.Space(obs=self.obs, limits=[l, u])
                return pdf.ext_integrate(limits_space, norm=False, options=options)

            missing_yield = False
        else:

            @z.function
            def integrate_one(limits):
                l, u = tf.unstack(limits)
                limits_space = zfit.Space(obs=self.obs, limits=[l, u])
                return pdf.integrate(limits_space, norm=False, options=options)

            missing_yield = True

        limits = znp.stack([lower_flat, upper_flat], axis=1)
        from zfit import run

        try:
            if run.executing_eagerly():
                raise TypeError("Just stearing the eager execution")
            values = tf.vectorized_map(integrate_one, limits)[:, 0]
        except (ValueError, TypeError):
            with run.aquire_cpu(-1) as cpus:
                values = tf.map_fn(integrate_one, limits, parallel_iterations=len(cpus))
        values = znp.reshape(values, shape)
        if missing_yield:
            values *= self.get_yield()
        if norm:
            values /= pdf.normalization(norm)
        return values



#  Copyright (c) 2023 zfit
import hist
import numpy as np

import zfit
import zfit.z.numpy as znp

# TODO: what is needed in this file?
from zfit import z
from zfit.core.binning import histogramdd, midpoints_from_hist, unbinned_to_binindex

data1 = np.random.normal(size=(1000, 3))
obs1 = zfit.Space("obs1", limits=(-100, 300))
obs2 = zfit.Space("obs2", limits=(-100, 300))
obs3 = zfit.Space("obs3", limits=(-100, 300))

obs = obs1 * obs2 * obs3


def test_histogramdd():
    histdd_kwargs = {"sample": data1}
    hist = histogramdd(**histdd_kwargs)
    bincount_np, edges_np = zfit.run(hist)
    bincount_true, edges_true = np.histogramdd(**histdd_kwargs)
    np.testing.assert_allclose(bincount_true, bincount_np)
    np.testing.assert_allclose(edges_true, edges_np)


def test_midpoints():
    edges = np.array([[-1.0, 0, 3, 10], [-5.0, 0, 1, 4]])
    bincounts = np.array([[0, 0, 1], [0, 5, 7], [0, 3, 0], [0, 0, 0]])

    edges = z.convert_to_tensor(edges)
    midpoints_true = np.array([[-0.5, 2.5], [1.5, 0.5], [1.5, 2.5], [6.5, 0.5]])
    bincounts_nonzero, midpoints_nonzero, bincounts_nonzero_index = midpoints_from_hist(
        bincounts=bincounts, edges=edges
    )
    np.testing.assert_allclose(np.array([1, 5, 7, 3]), zfit.run(bincounts_nonzero))
    np.testing.assert_allclose(midpoints_true, zfit.run(midpoints_nonzero))


def test_unbinned_to_bins():
    n = 100_000
    lower = [-5, -50, 10]
    upper = [5, 13, 100]
    values = np.random.uniform(lower, upper, size=(n, 3))
    axes = [
        hist.axis.Regular(12, lower[0], upper[0], name="x"),
        hist.axis.Regular(24, lower[1], upper[1], name="y"),
        hist.axis.Variable(
            [10, 20, 23, 26.5, 30, 35.0, 50, 60.0, 75, 78, 90.0, 100], name="z"
        ),
    ]
    h = hist.NamedHist(*axes)
    name_values = dict(zip(["x", "y", "z"], values.transpose()))
    # h.fill(**name_values)
    true_bins = h.axes.index(*name_values.values())
    space = zfit.Space(binning=axes)
    data = zfit.Data.from_tensor(space.with_binning(None), values)
    bins = znp.transpose(unbinned_to_binindex(data, space))
    np.testing.assert_array_equal(bins, true_bins)



#  Copyright (c) 2023 zfit
from __future__ import annotations

import zfit.z.numpy as znp
from .functor import BaseFunctor
from ..core.interfaces import ZfitBinnedPDF
from ..core.space import supports
from ..util import ztyping
from ..util.exception import SpecificFunctionNotImplemented
from ..util.ztyping import ExtendedInputType, NormInputType
from ..z.interpolate_spline import interpolate_spline


class SplinePDF(BaseFunctor):
    def __init__(
        self,
        pdf: ZfitBinnedPDF,
        order: int | None = None,
        obs: ztyping.ObsTypeInput = None,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
    ) -> None:
        """Spline interpolate a binned PDF in order to get a smooth, unbinned PDF.

        Args:
            pdf: Binned PDF that will be interpolated.
            order: Spline interpolation order. Default is 3
            obs: Unbinned observable. If not given, the observable of the pdf is used without the binning.
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| The normalization of the PDF. If this is parameter-like, it will be used as the
        """
        if extended is None:
            extended = pdf.is_extended
        if extended is True:
            extended = pdf.get_yield()
            self._automatically_extended = True
        else:
            self._automatically_extended = False
        if obs is None:
            obs = pdf.space
            obs = obs.with_binning(None)
        super().__init__(pdfs=pdf, obs=obs, extended=extended, norm=norm)
        if order is None:
            order = 3
        self._order = order

    @property
    def order(self):
        return self._order

    @supports(norm=True)
    def _ext_pdf(self, x, norm):
        if not self._automatically_extended:
            raise SpecificFunctionNotImplemented
        pdf = self.pdfs[0]
        density = pdf.ext_pdf(x.space, norm=norm)
        density_flat = znp.reshape(density, (-1,))
        centers_list = znp.meshgrid(*pdf.space.binning.centers, indexing="ij")
        centers_list_flat = [znp.reshape(cent, (-1,)) for cent in centers_list]
        centers = znp.stack(centers_list_flat, axis=-1)
        # [None, :, None]  # TODO: only 1 dim now
        probs = interpolate_spline(
            train_points=centers[None, ...],
            train_values=density_flat[None, :, None],
            query_points=x.value()[None, ...],
            order=self.order,
        )
        return probs[0, ..., 0]

    @supports(norm=True)
    def _pdf(self, x, norm):
        pdf = self.pdfs[0]
        density = pdf.pdf(
            x.space, norm=norm
        )  # TODO: order? Give obs, pdf makes order and binning herself?
        centers = pdf.space.binning.centers[0][None, :, None]  # TODO: only 1 dim now
        probs = interpolate_spline(
            train_points=centers,
            train_values=density[None, :, None],
            query_points=x.value()[None, ...],
            order=3,
        )
        return probs[0, ..., 0]



#  Copyright (c) 2024 zfit

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import zfit

from collections.abc import Mapping

from .binned_functor import BaseBinnedFunctorPDF
from ..core.space import supports
from ..core.interfaces import ZfitBinnedPDF
import tensorflow as tf
import zfit.z.numpy as znp
from ..util import ztyping
from ..util.exception import SpecificFunctionNotImplemented


class BinwiseScaleModifier(BaseBinnedFunctorPDF):
    def __init__(
        self,
        pdf: ZfitBinnedPDF,
        modifiers: bool | Mapping[str, ztyping.ParamTypeInput] = None,
        extended: ztyping.ExtendedInputType = None,
        norm: ztyping.NormInputType = None,
        name: str | None = "BinnedTemplatePDF",
    ) -> None:
        """Modifier that scales each bin separately of the *pdf*.

        Binwise modification can be used to account for uncorrelated or correlated uncertainties.

        Args:
            pdf: Binned pdf to be modified.
            modifiers: Modifiers for each bin.
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: |@doc:model.init.name| Human-readable name
               or label of
               the PDF for better identification.
               Has no programmatical functional purpose as identification. |@docend:model.init.name|
        """
        obs = pdf.space
        if not isinstance(pdf, ZfitBinnedPDF):
            raise TypeError("pdf must be a BinnedPDF")
        if extended is None:
            extended = pdf.is_extended
        if modifiers is None:
            modifiers = True
        if modifiers is True:
            import zfit

            modifiers = {
                f"sysshape_{i}": zfit.Parameter(f"auto_sysshape_{self}_{i}", 1.0)
                for i in range(pdf.counts(obs).shape.num_elements())
            }
        if not isinstance(modifiers, dict):
            raise TypeError("modifiers must be a dict-like object or True or None")
        params = modifiers.copy()
        self._binwise_modifiers = modifiers
        if extended is True:
            self._automatically_extended = True
            if modifiers:
                import zfit

                def sumfunc(params):
                    values = self.counts()
                    return znp.sum(values)

                from zfit.core.parameter import get_auto_number

                params_sumfunc = modifiers.copy()
                dep_params = {}
                for p in pdf.get_params():
                    while True:
                        i = get_auto_number()
                        name_tmp = f"DUMMPY_PARAM_{i}"
                        if name_tmp not in params_sumfunc:
                            dep_params[name_tmp] = p
                            break

                params_sumfunc.update(dep_params)
                extended = zfit.ComposedParameter(
                    f"AUTO_binwise_modifier_{get_auto_number()}",
                    sumfunc,
                    params=params_sumfunc,
                )

            else:
                extended = pdf.get_yield()
        elif extended is not False:
            self._automatically_extended = False
        super().__init__(
            obs=obs, name=name, params=params, models=pdf, extended=extended, norm=norm
        )

    @supports(norm=True)
    def _counts(self, x, norm=None):
        if not self._automatically_extended:
            raise SpecificFunctionNotImplemented
        values = self._counts_with_modifiers(x, norm)
        return values

    def _counts_with_modifiers(self, x, norm):
        values = self.pdfs[0].counts(x, norm=norm)
        if modifiers := list(self._binwise_modifiers.values()):
            sysshape_flat = tf.stack(modifiers)
            modifiers = znp.reshape(sysshape_flat, values.shape)
            values = values * modifiers
        return values

    @supports(norm="space")
    def _rel_counts(self, x, norm=None):
        values = self._counts_with_modifiers(x, norm)
        return values / znp.sum(values)



"""Top-level package for zfit."""

#  Copyright (c) 2023 zfit
import sys
import warnings

if sys.version_info < (3, 8):
    from importlib_metadata import version
else:
    from importlib.metadata import version

__version__ = version(__name__)

__license__ = "BSD 3-Clause"
__copyright__ = "Copyright 2018, zfit"
__status__ = "Beta"

__author__ = (
    "Jonas Eschle <Jonas.Eschle@cern.ch>,"
    "Albert Puig <albert@protonmail.com>, "
    "Rafael Silva Coutinho <rsilvaco@cern.ch>, "
    "Matthieu Marinangeli <matthieu.marinangeli@cern.ch>"
)
__maintainer__ = "zfit"
__email__ = "zfit@physik.uzh.ch"
__credits__ = (
    "Chris Burr, Martina Ferrillo, Abhijit Mathad, Oliver Lantwin, Johannes Lade"
)

__all__ = [
    "z",
    "constraint",
    "pdf",
    "minimize",
    "loss",
    "data",
    "func",
    "dimension",
    "exception",
    "sample",
    "binned",
    "hs3",
    "Parameter",
    "ComposedParameter",
    "ComplexParameter",
    "convert_to_parameter",
    "Space",
    "convert_to_space",
    "supports",
    "run",
    "settings",
]


#  Copyright (c) 2019 zfit


def _maybe_disable_warnings():
    import os

    disable_warnings = os.environ.get("ZFIT_DISABLE_TF_WARNINGS")
    if disable_warnings is None:
        warnings.warn(
            "TensorFlow warnings are by default suppressed by zfit."
            " In order to show them,"
            " set the environment variable ZFIT_DISABLE_TF_WARNINGS=0."
            " In order to suppress the TensorFlow warnings AND this warning,"
            " set ZFIT_DISABLE_TF_WARNINGS=1."
        )
    elif disable_warnings == "0":  # just ignore and do nothing
        return

    os.environ["KMP_AFFINITY"] = "noverbose"
    os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"

    import tensorflow as tf

    tf.get_logger().setLevel("ERROR")


_maybe_disable_warnings()

import tensorflow as tf

if int(tf.__version__[0]) < 2:
    raise RuntimeError(
        f"You are using TensorFlow version {tf.__version__}. This zfit version ({__version__}) works"
        f" only with TF >= 2"
    )

from . import z  # initialize first
from . import (
    constraint,
    data,
    dimension,
    exception,
    func,
    loss,
    binned,
    minimize,
    param,
    pdf,
    sample,
    settings,
    hs3,
)
from .core.data import Data
from .core.parameter import (
    ComplexParameter,
    ComposedParameter,
    Parameter,
    convert_to_parameter,
)
from .core.space import Space, convert_to_space, supports
from .settings import run, ztypes
from .util.graph import jit as _jit


def _maybe_disable_jit():
    import os

    arg1 = os.environ.get("ZFIT_DO_JIT")
    arg2 = os.environ.get("ZFIT_EXPERIMENTAL_DO_JIT")
    arg3 = os.environ.get("ZFIT_MODE_GRAPH")
    if arg3 is not None:
        warnings.warn(
            "Depreceated to use `ZFIT_MODE_GRAPH`, use `ZFIT_GRAPH_MODE` instead.",
            DeprecationWarning,
        )

    if arg1 is not None and arg2 is None:
        warnings.warn(
            "Depreceated to use `ZFIT_EXPERIMENTAL_DO_JIT`, use `ZFIT_GRAPH_MODE` instead.",
            DeprecationWarning,
        )
    arg = arg2 if arg1 is None else arg1
    if arg is not None and not int(arg):
        run.set_graph_mode(False)

    graph = os.environ.get("ZFIT_GRAPH_MODE")
    if graph is not None and not int(graph):
        run.set_graph_mode(False)


# experimental flags


_maybe_disable_jit()

# EOF



#  Copyright (c) 2022 zfit

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..models.functor import ProductPDF, SumPDF
    from ..models.functions import ProdFunc, SumFunc

from collections.abc import Callable

import zfit.z.numpy as znp

from ..util import ztyping
from ..util.exception import (
    BreakingAPIChangeError,
    IntentionAmbiguousError,
    ModelIncompatibleError,
)
from .interfaces import ZfitFunc, ZfitModel, ZfitParameter, ZfitPDF
from .parameter import convert_to_parameter


def multiply(
    object1: ztyping.BaseObjectType, object2: ztyping.BaseObjectType
) -> ztyping.BaseObjectType:
    """Multiply two objects and return a new object (may depending on the old).

    Args:
        object1: A ZfitParameter, ZfitFunc or ZfitPDF to multiply with object2
        object2: A ZfitParameter, ZfitFunc or ZfitPDF to multiply with object1
    Raises:
        TypeError: if one of the objects is neither a ZfitFunc, ZfitPDF or convertable to a ZfitParameter
    """
    # converting the objects to known types
    object1, object2 = _convert_to_known(object1, object2)
    new_object = None

    # object 1 is ZfitParameter
    if isinstance(object1, ZfitParameter):
        if isinstance(object2, ZfitParameter):
            new_object = multiply_param_param(param1=object1, param2=object2)
        elif isinstance(object2, ZfitFunc):
            new_object = multiply_param_func(param=object1, func=object2)
        elif isinstance(object2, ZfitPDF):
            new_object = multiply_param_pdf(param=object1, pdf=object2)
        else:
            assert (
                False
            ), "This code should never be reached due to logical reasons. Mistakes happen..."

    # object 1 is Function
    elif isinstance(object1, ZfitFunc):
        if isinstance(object2, ZfitParameter):
            new_object = multiply_param_func(param=object2, func=object1)
        elif isinstance(object2, ZfitFunc):
            new_object = multiply_func_func(func1=object1, func2=object2)
        elif isinstance(object2, ZfitPDF):
            raise ModelIncompatibleError(
                "Cannot multiply a function with a model. Use `func.as_pdf` or `model.as_func`."
            )

    # object 1 is PDF
    elif isinstance(object1, ZfitPDF):
        if isinstance(object2, ZfitPDF):
            new_object = multiply_pdf_pdf(pdf1=object1, pdf2=object2)

    if new_object is None:
        raise ModelIncompatibleError(
            "Multiplication for {} and {} of type {} and {} is not"
            "properly defined. (may change the order)"
            "".format(object1, object2, type(object1), type(object2))
        )
    return new_object


def multiply_pdf_pdf(
    pdf1: ZfitPDF, pdf2: ZfitPDF, name: str = "multiply_pdf_pdf"
) -> ProductPDF:
    if not (isinstance(pdf1, ZfitPDF) and isinstance(pdf2, ZfitPDF)):
        raise TypeError(
            f"`pdf1` and `pdf2` need to be `ZfitPDF` and not {pdf1}, {pdf2}"
        )
    from ..models.functor import ProductPDF

    if not pdf1.is_extended and pdf2.is_extended:
        raise IntentionAmbiguousError(
            "Cannot multiply this way a non-extendended PDF with an extended PDF."
            "Only vice-versa is allowed: to multiply an extended PDF with an "
            "non-extended PDF."
        )

    return ProductPDF(pdfs=[pdf1, pdf2], name=name)


def multiply_func_func(
    func1: ZfitFunc, func2: ZfitFunc, name: str = "multiply_func_func"
) -> ProdFunc:
    if not (isinstance(func1, ZfitFunc) and isinstance(func2, ZfitFunc)):
        raise TypeError(
            f"`func1` and `func2` need to be `ZfitFunc` and not {func1}, {func2}"
        )
    from ..models.functions import ProdFunc

    return ProdFunc(funcs=[func1, func2], name=name)


def multiply_param_pdf(param: ZfitParameter, pdf: ZfitPDF) -> ZfitPDF:
    # TODO(SUM): remove it completely?
    raise BreakingAPIChangeError(
        "Unfortunately, it is not allowed anymore to multiply pdfs with `frac * pdf + other_pdf` syntax"
        "due to ambiguity. Use the `zfit.pdf.SumPDF([pdf, other_pdf], frac)` syntax instead."
    )

    # if not (isinstance(param, ZfitParameter) and isinstance(pdf, ZfitPDF)):
    #     raise TypeError("`param` and `model` need to be `ZfitParameter` resp. `ZfitPDF` and not "
    #                     "{}, {}".format(param, pdf))
    # if pdf.is_extended:
    #     raise AlreadyExtendedPDFError()
    #
    # new_pdf = pdf.create_extended(param, name_addition="_autoextended")
    # return new_pdf


def multiply_param_func(param: ZfitParameter, func: ZfitFunc) -> ZfitFunc:
    if not (isinstance(param, ZfitParameter) and isinstance(func, ZfitFunc)):
        raise TypeError(
            "`param` and `func` need to be `ZfitParameter` resp. `ZfitFunc` and not "
            "{}, {}".format(param, func)
        )
    from ..models.functions import SimpleFuncV1

    def combined_func(x):
        return param * func.func(x=x)

    params = {param.name: param}
    params.update(func.params)
    new_func = SimpleFuncV1(
        func=combined_func, obs=func.obs, **params
    )  # TODO: implement with new parameters
    return new_func


def multiply_param_param(param1: ZfitParameter, param2: ZfitParameter) -> ZfitParameter:
    if not (isinstance(param1, ZfitParameter) and isinstance(param2, ZfitParameter)):
        raise TypeError(
            f"`param1` and `param2` need to be `ZfitParameter` and not {param1}, {param2}"
        )
    param = znp.multiply(param1, param2)
    return param


# Addition logic
def add(
    object1: ztyping.BaseObjectType, object2: ztyping.BaseObjectType
) -> ztyping.BaseObjectType:
    """Add two objects and return a new object (may depending on the old).

    Args:
        object1: A ZfitParameter, ZfitFunc or ZfitPDF to add with object2
        object2: A ZfitParameter, ZfitFunc or ZfitPDF to add with object1
    """
    # converting the objects to known types
    object1, object2 = _convert_to_known(object1, object2)
    new_object = None

    # object 1 is ZfitParameter
    if isinstance(object1, ZfitParameter):
        if isinstance(object2, ZfitParameter):
            new_object = add_param_param(param1=object1, param2=object2)
        elif isinstance(object2, ZfitFunc):
            new_object = add_param_func(param=object1, func=object2)

    # object 1 is Function
    elif isinstance(object1, ZfitFunc):
        if isinstance(object2, ZfitParameter):
            new_object = add_param_func(param=object2, func=object1)
        elif isinstance(object2, ZfitFunc):
            new_object = add_func_func(func1=object1, func2=object2)
        elif isinstance(object2, ZfitPDF):
            raise TypeError(
                "Cannot add a function with a model. Use `func.as_pdf` or `model.as_func`."
            )

    # object 1 is PDF
    elif isinstance(object1, ZfitPDF):
        if isinstance(object2, ZfitFunc):
            raise TypeError(
                "Cannot add a function with a model. Use `func.as_pdf` or `model.as_func`."
            )
        elif isinstance(object2, ZfitPDF):
            new_object = add_pdf_pdf(pdf1=object1, pdf2=object2)

    if new_object is None:
        raise TypeError(
            "Addition for {} and {} of type {} and {} is not"
            "properly defined. (may change the order)"
            "".format(object1, object2, type(object1), type(object2))
        )
    return new_object


def _convert_to_known(object1, object2):
    objects = []
    for obj in (object1, object2):
        if not isinstance(obj, (ZfitModel,)):
            try:
                obj = convert_to_parameter(obj)
            except TypeError as error:
                raise TypeError(
                    "Object is neither an instance of ZfitModel nor convertible to a Parameter: "
                    "{}".format(obj)
                )
        objects.append(obj)
    object1, object2 = objects
    return object1, object2


def add_pdf_pdf(pdf1: ZfitPDF, pdf2: ZfitPDF, name: str = "add_pdf_pdf") -> SumPDF:
    if not (isinstance(pdf1, ZfitPDF) and isinstance(pdf2, ZfitPDF)):
        raise TypeError(
            f"`pdf1` and `pdf2` need to be `ZfitPDF` and not {pdf1}, {pdf2}"
        )
    if not (pdf1.is_extended and pdf2.is_extended):
        raise BreakingAPIChangeError(
            "Adding (non-extended) pdfs is not allowed anymore due to disambiguity."
            "Use the `zfit.pdf.SumPDF([pdf, other_pdf], frac)` syntax instead."
        )
    from ..models.functor import SumPDF

    return SumPDF(pdfs=[pdf1, pdf2], name=name)


def add_func_func(
    func1: ZfitFunc, func2: ZfitFunc, name: str = "add_func_func"
) -> SumFunc:
    if not (isinstance(func1, ZfitFunc) and isinstance(func2, ZfitFunc)):
        raise TypeError(
            f"`func1` and `func2` need to be `ZfitFunc` and not {func1}, {func2}"
        )
    from ..models.functions import SumFunc

    return SumFunc(funcs=[func1, func2], name=name)


def add_param_func(param: ZfitParameter, func: ZfitFunc) -> ZfitFunc:
    if not (isinstance(param, ZfitParameter) and isinstance(func, ZfitFunc)):
        raise TypeError(
            "`param` and `func` need to be `ZfitParameter` resp. `ZfitFunc` and not "
            "{}, {}".format(param, func)
        )
    raise NotImplementedError(
        "This is not supported. Probably in the future."
    )  # TODO: implement with new parameters


def add_param_param(param1: ZfitParameter, param2: ZfitParameter) -> ZfitParameter:
    if not (isinstance(param1, ZfitParameter) and isinstance(param2, ZfitParameter)):
        raise TypeError(
            f"`param1` and `param2` need to be `ZfitParameter` and not {param1}, {param2}"
        )
    # use the default behavior of variables
    return znp.add(param1, param2)


# Conversions


def convert_pdf_to_func(pdf: ZfitPDF, norm: ztyping.LimitsType) -> ZfitFunc:
    def value_func(x):
        return pdf.pdf(x, norm=norm)

    from ..models.functions import SimpleFuncV1

    func = SimpleFuncV1(
        func=value_func, obs=pdf.obs, name=pdf.name + "_as_func", **pdf.params
    )
    return func


def convert_func_to_pdf(func: ZfitFunc | Callable, obs=None, name=None) -> ZfitPDF:
    func_name = "autoconverted_func_to_pdf" if name is None else name
    if not isinstance(func, ZfitFunc) and callable(func):
        if obs is None:
            raise ValueError("If `func` is a function, `obs` has to be specified.")
        from ..models.functions import SimpleFuncV1

        func = SimpleFuncV1(func=func, obs=obs, name=func_name)
    from ..models.special import SimplePDF

    name = func.name if name is None else func_name
    pdf = SimplePDF(func=func.func, obs=func.obs, name=name, **func.params)
    return pdf



#  Copyright (c) 2023 zfit
from __future__ import annotations

import numpy as np
from uhi.typing.plottable import PlottableHistogram

import zfit.z.numpy as znp
from ..core.binnedpdf import BaseBinnedPDFV1
from ..core.interfaces import ZfitBinnedData
from ..core.space import supports
from ..util import ztyping
from ..util.exception import SpecificFunctionNotImplemented


class HistogramPDF(BaseBinnedPDFV1):
    def __init__(
        self,
        data: ztyping.BinnedDataInputType,
        extended: ztyping.ExtendedInputType | None = None,
        norm: ztyping.NormInputType | None = None,
        name: str = "HistogramPDF",
    ) -> None:
        """Binned PDF resembling a histogram.

        Simple histogram PDF that can be used to model a histogram as a PDF.


        Args:
            data: Histogram to be used as PDF.
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
               |@doc:pdf.init.extended.auto| If ``True``,
               the PDF will be extended automatically if the PDF is extended
               using the total number of events in the histogram.
               This is the default. |@docend:pdf.init.extended.auto|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: |@doc:model.init.name| Human-readable name
               or label of
               the PDF for better identification.
               Has no programmatical functional purpose as identification. |@docend:model.init.name|
        """
        if extended is None:
            extended = True
        if not isinstance(data, ZfitBinnedData):
            if isinstance(data, PlottableHistogram):
                from zfit._data.binneddatav1 import BinnedData

                data = BinnedData.from_hist(data)
            else:
                raise TypeError(
                    "data must be of type PlottableHistogram (UHI) or ZfitBinnedData"
                )

        params = {}
        if extended is True:
            self._automatically_extended = True
            extended = znp.sum(data.values())
        else:
            self._automatically_extended = False
        super().__init__(
            obs=data.space, extended=extended, norm=norm, params=params, name=name
        )
        self._data = data

    @supports(norm="space")
    def _ext_pdf(self, x, norm):
        if not self._automatically_extended:
            raise SpecificFunctionNotImplemented
        counts = self._counts(x, norm)
        areas = np.prod(self._data.axes.widths, axis=0)
        density = counts / areas
        return density

    @supports(norm="space")
    def _pdf(self, x, norm):
        counts = self._rel_counts(x, norm)
        areas = np.prod(self._data.axes.widths, axis=0)
        density = counts / areas
        return density

    @supports(norm="space")
    def _counts(self, x, norm=None):
        if not self._automatically_extended:
            raise SpecificFunctionNotImplemented
        values = self._data.values()
        return values

    @supports(norm="space")
    def _rel_counts(self, x, norm=None):
        values = self._data.values()
        return values / znp.sum(values)
